from __future__ import annotations
from typing import Any
from ...arctrl_core.arc_types import ArcRun
from ...arctrl_core.comment import Comment
from ...arctrl_core.data_map import DataMap
from ...arctrl_core.Helper.collections_ import (ResizeArray_singleton, ResizeArray_map, ResizeArray_zip, ResizeArray_appendSingleton, Option_fromSeq, Option_fromValueWithDefault, ResizeArray_filter)
from ...arctrl_core.Helper.identifier import Run_cwlFileNameFromIdentifier
from ...arctrl_core.ontology_annotation import OntologyAnnotation
from ...arctrl_core.person import Person
from ...arctrl_core.Table.arc_tables import ArcTables
from ...arctrl_cwl.cwlprocessing_unit import CWLProcessingUnit
from ...arctrl_cwl.cwltypes import (CWLType, CWLType_file)
from ...arctrl_cwl.inputs import (InputBinding, CWLInput)
from ...arctrl_cwl.parameter_reference import CWLParameterReference
from ...arctrl_file_system.file_system import FileSystem
from ...arctrl_file_system.path import (combine_many, combine)
from ...arctrl_rocrate.ldcontext import LDContext
from ...arctrl_rocrate.ldobject import (LDNode, LDGraph)
from ...arctrl_rocrate.LDTypes.computational_workflow import LDComputationalWorkflow
from ...arctrl_rocrate.LDTypes.dataset import LDDataset
from ...arctrl_rocrate.LDTypes.file import LDFile
from ...arctrl_rocrate.LDTypes.formal_parameter import LDFormalParameter
from ...arctrl_rocrate.LDTypes.lab_process import LDLabProcess
from ...arctrl_rocrate.LDTypes.property_value import LDPropertyValue
from ...arctrl_rocrate.LDTypes.workflow_invocation import LDWorkflowInvocation
from ...fable_library.list import of_seq
from ...fable_library.option import (value as value_1, default_arg, bind, map)
from ...fable_library.reflection import (TypeInfo, class_type)
from ...fable_library.seq import (try_pick, find, exists)
from ...fable_library.string_ import (starts_with_exact, replace, trim_start, join, to_fail, printf)
from ...fable_library.types import (to_string, Array)
from ...fable_library.util import equals
from .assay import AssayConversion_getDataFilesFromProcesses_6BABD1B0
from .basic import (BaseTypes_composeDefinedTerm_ZDED3A0F, BaseTypes_composePropertyValueFromOA_ZDED3A0F, BaseTypes_composeComment_Z13201A7E, BaseTypes_decomposeDefinedTerm_Z2F770004, BaseTypes_decomposePropertyValueToOA_Z2F770004, BaseTypes_decomposeComment_Z2F770004)
from .datamap import (DatamapConversion_composeFragmentDescriptors_Z8923FA3, DatamapConversion_decomposeFragmentDescriptors_Z6E59645F)
from .person import (PersonConversion_composePerson_Z64D846DC, PersonConversion_decomposePerson_Z6839B9E8)
from .table import (ARCtrl_ArcTables__ArcTables_GetProcesses_5E660E5C, ARCtrl_ArcTables__ArcTables_fromProcesses_Static_Z27F0B586)
from .workflow import (WorkflowConversion_composeWorkflowProtocolFromProcessingUnit_Z3099C0F7, WorkflowConversion_getInputParametersFromProcessingUnit_30922B92, WorkflowConversion_decomposeWorkflowProtocolToProcessingUnit_2664C4B8)

def _expr3977() -> TypeInfo:
    return class_type("ARCtrl.Conversion.RunConversion", None, RunConversion)


class RunConversion:
    ...

RunConversion_reflection = _expr3977

def RunConversion_composeCWLInputFilePath_Z384F8060(path: str, run_name: str) -> str:
    if starts_with_exact(path, "../.."):
        return replace(replace(path, "../../", ""), "../..", "")

    else: 
        return combine_many(["runs", run_name, path])



def RunConversion_decomposeCWLInputFilePath_Z384F8060(path: str, run_name: str) -> str:
    prefix: str = combine_many(["runs", run_name])
    if starts_with_exact(path, prefix):
        return trim_start(replace(path, prefix, ""), "/")

    else: 
        return combine("../..", path)



def RunConversion_composeCWLInputValue_70DD9184(input_value: CWLParameterReference, example_of_work: LDNode, input_param: CWLInput, run_name: str) -> LDNode:
    if input_param.Type_ is None:
        raise Exception(((("Cannot convert param values \"" + str(input_value.Values)) + "\" as Input parameter \"") + input_param.Name) + "\" has no type.")

    type_: CWLType = value_1(input_param.Type_)
    if input_value.Type is not None:
        if not equals(value_1(input_value.Type), type_):
            raise Exception(((((("Type (" + to_string(value_1(input_value.Type))) + ") of yml input value \"") + input_value.Key) + "\" does not match type of workflow input parameter (") + to_string(type_)) + ").")


    (pattern_matching_result,) = (None,)
    if type_.tag == 0:
        if len(input_value.Values) == 1:
            pattern_matching_result = 0

        elif to_string(type_).lower().find("array") >= 0:
            pattern_matching_result = 1

        else: 
            pattern_matching_result = 2


    elif to_string(type_).lower().find("array") >= 0:
        pattern_matching_result = 1

    else: 
        pattern_matching_result = 2

    if pattern_matching_result == 0:
        path: str = RunConversion_composeCWLInputFilePath_Z384F8060(input_value.Values[0], run_name)
        return LDFile.create_cwlparameter(path, example_of_work)

    elif pattern_matching_result == 1:
        def binder(ib: InputBinding, input_value: Any=input_value, example_of_work: Any=example_of_work, input_param: Any=input_param, run_name: Any=run_name) -> str | None:
            return ib.ItemSeparator

        values: str = join(default_arg(bind(binder, input_param.InputBinding), ","), input_value.Values)
        return LDPropertyValue.create_cwlparameter(example_of_work, input_value.Key, ResizeArray_singleton(values))

    elif pattern_matching_result == 2:
        return LDPropertyValue.create_cwlparameter(example_of_work, input_value.Key, input_value.Values)



def RunConversion_decomposeCWLInputValue_Z1346FE9D(input_value: LDNode, run_name: str, context: LDContext | None=None, graph: LDGraph | None=None) -> CWLParameterReference:
    example_of_work: LDNode
    match_value: LDNode | None = LDFile.try_get_example_of_work_as_formal_parameter(input_value, graph, context)
    if match_value is None:
        arg: str = input_value.Id
        example_of_work = to_fail(printf("Input value %s of run %s must have an exampleOfWork property pointing to a CWL formal parameter."))(arg)(run_name)

    else: 
        example_of_work = match_value

    key: str = LDFormalParameter.get_name_as_string(example_of_work, context)
    if LDFile.validate_cwlparameter(input_value, context):
        return CWLParameterReference(key, [RunConversion_decomposeCWLInputFilePath_Z384F8060(input_value.Id, run_name)], CWLType_file())

    elif LDPropertyValue.validate_cwlparameter(input_value, context):
        return CWLParameterReference(key, LDPropertyValue.get_values_as_string(input_value, context))

    else: 
        arg_2: str = input_value.Id
        return to_fail(printf("Input value %s of run %s is neither a CWL File nor a CWL Parameter."))(arg_2)(run_name)



def RunConversion_composeWorkflowInvocationFromArcRun_Z8CC08AC(run: ArcRun, fs: FileSystem | None=None) -> Array[LDNode]:
    workflow_protocol: LDNode
    workflow_file_path: str = Run_cwlFileNameFromIdentifier(run.Identifier)
    match_value: CWLProcessingUnit | None = run.CWLDescription
    if match_value is None:
        arg: str = run.Identifier
        workflow_protocol = to_fail(printf("Run %s must have a CWL description"))(arg)

    else: 
        workflow_protocol = WorkflowConversion_composeWorkflowProtocolFromProcessingUnit_Z3099C0F7(workflow_file_path, match_value, None, run.Identifier)

    def f(tupled_arg: tuple[CWLInput, LDNode], run: Any=run, fs: Any=fs) -> LDNode:
        ld_i: LDNode = tupled_arg[1]
        name: str = LDFormalParameter.get_name_as_string(ld_i, workflow_protocol.TryGetContext())
        def chooser(i_1: CWLParameterReference, tupled_arg: Any=tupled_arg) -> CWLParameterReference | None:
            if i_1.Key == name:
                return i_1

            else: 
                return None


        param_ref: CWLParameterReference | None = try_pick(chooser, run.CWLInput)
        if param_ref is None:
            raise Exception(("Could not create workflow invocation for run \"" + run.Identifier) + "\": Workflow parameter \"name\" had no assigned value.")

        else: 
            return RunConversion_composeCWLInputValue_70DD9184(param_ref, ld_i, tupled_arg[0], run.Identifier)


    def _arrow3978(__unit: None=None, run: Any=run, fs: Any=fs) -> Array[tuple[CWLInput, LDNode]]:
        b: Array[LDNode] = LDComputationalWorkflow.get_inputs_as_formal_parameters(workflow_protocol, None, workflow_protocol.TryGetContext())
        return ResizeArray_zip(WorkflowConversion_getInputParametersFromProcessingUnit_30922B92(value_1(run.CWLDescription)), b)

    input_params: Array[LDNode] = ResizeArray_map(f, _arrow3978())
    process_sequence: Array[LDNode] = list(ARCtrl_ArcTables__ArcTables_GetProcesses_5E660E5C(ArcTables(run.Tables), None, None, fs))
    main_invocation: LDNode = LDWorkflowInvocation.create(run.Identifier, workflow_protocol, input_params, None, None, None, None, None, None, workflow_protocol)
    if len(process_sequence) == 0:
        return ResizeArray_singleton(main_invocation)

    else: 
        def f_1(p: LDNode, run: Any=run, fs: Any=fs) -> LDNode:
            id: str = replace(p.Id, "Process", ("WorkflowInvocation_" + run.Identifier) + "")
            name_1: str = LDLabProcess.get_name_as_string(p, p.TryGetContext())
            inputs: Array[LDNode] = LDLabProcess.get_objects(p)
            results: Array[LDNode] | None = Option_fromSeq(LDLabProcess.get_results(p))
            protocol: LDNode | None = LDLabProcess.try_get_executes_lab_protocol(p)
            parameter_values: Array[LDNode] | None = Option_fromSeq(LDLabProcess.get_parameter_values(p))
            def mapping(a_3: LDNode, p: Any=p) -> Array[LDNode]:
                return ResizeArray_singleton(a_3)

            agents: Array[LDNode] | None = map(mapping, LDLabProcess.try_get_agent(p))
            disambiguating_descriptions: Array[str] | None = Option_fromSeq(LDLabProcess.get_disambiguating_descriptions_as_string(p))
            return LDWorkflowInvocation.create(name_1, workflow_protocol, inputs, results, None, agents, id, None, disambiguating_descriptions, protocol, parameter_values)

        return ResizeArray_appendSingleton(main_invocation, ResizeArray_map(f_1, process_sequence))



def RunConversion_decomposeMainWorkflowInvocation_Z1346FE9D(workflow_invocation: LDNode, run_name: str, context: LDContext | None=None, graph: LDGraph | None=None) -> tuple[CWLProcessingUnit, Array[CWLParameterReference]]:
    def _arrow3979(__unit: None=None, workflow_invocation: Any=workflow_invocation, run_name: Any=run_name, context: Any=context, graph: Any=graph) -> CWLProcessingUnit:
        match_value: LDNode | None = LDLabProcess.try_get_executes_lab_protocol(workflow_invocation, graph, context)
        if match_value is None:
            raise Exception(("Could not decompose workflow invocation for run \"" + run_name) + "\": Workflow parameter \"name\" had no assigned value.")

        else: 
            return WorkflowConversion_decomposeWorkflowProtocolToProcessingUnit_2664C4B8(match_value, context, graph)


    def f(iv: LDNode, workflow_invocation: Any=workflow_invocation, run_name: Any=run_name, context: Any=context, graph: Any=graph) -> CWLParameterReference:
        return RunConversion_decomposeCWLInputValue_Z1346FE9D(iv, run_name, context, graph)

    return (_arrow3979(), ResizeArray_map(f, LDLabProcess.get_objects(workflow_invocation, graph, context)))


def RunConversion_composeRun_Z8CC08AC(run: ArcRun, fs: FileSystem | None=None) -> LDNode:
    workflow_invocations: Array[LDNode] | None = Option_fromSeq(RunConversion_composeWorkflowInvocationFromArcRun_Z8CC08AC(run, fs))
    def mapping(term: OntologyAnnotation, run: Any=run, fs: Any=fs) -> LDNode:
        return BaseTypes_composeDefinedTerm_ZDED3A0F(term)

    measurement_method: LDNode | None = map(mapping, run.TechnologyType)
    def mapping_1(term_1: OntologyAnnotation, run: Any=run, fs: Any=fs) -> LDNode:
        return BaseTypes_composeDefinedTerm_ZDED3A0F(term_1)

    measurement_technique: LDNode | None = map(mapping_1, run.TechnologyPlatform)
    def mapping_2(term_2: OntologyAnnotation, run: Any=run, fs: Any=fs) -> LDNode:
        return BaseTypes_composePropertyValueFromOA_ZDED3A0F(term_2)

    variable_measured: LDNode | None = map(mapping_2, run.MeasurementType)
    def f(c: Person, run: Any=run, fs: Any=fs) -> LDNode:
        return PersonConversion_composePerson_Z64D846DC(c)

    creators: Array[LDNode] | None = Option_fromSeq(ResizeArray_map(f, run.Performers))
    def mapping_3(datamap: DataMap, run: Any=run, fs: Any=fs) -> Array[LDNode]:
        return DatamapConversion_composeFragmentDescriptors_Z8923FA3(datamap)

    fragment_descriptors: Array[LDNode] | None = map(mapping_3, run.DataMap)
    def mapping_4(ps: Array[LDNode], run: Any=run, fs: Any=fs) -> Array[LDNode]:
        return AssayConversion_getDataFilesFromProcesses_6BABD1B0(ps, fragment_descriptors)

    data_files: Array[LDNode] | None = map(mapping_4, workflow_invocations)
    def _arrow3980(__unit: None=None, run: Any=run, fs: Any=fs) -> Array[LDNode] | None:
        fds_1: Array[LDNode] = fragment_descriptors
        return fds_1

    def _arrow3981(__unit: None=None, run: Any=run, fs: Any=fs) -> Array[LDNode] | None:
        vm_1: LDNode = variable_measured
        return ResizeArray_singleton(vm_1)

    def _arrow3982(__unit: None=None, run: Any=run, fs: Any=fs) -> Array[LDNode] | None:
        fds: Array[LDNode] = fragment_descriptors
        vm: LDNode = variable_measured
        return ResizeArray_appendSingleton(vm, fds)

    variable_measureds: Array[LDNode] | None = (None if (fragment_descriptors is None) else _arrow3980()) if (variable_measured is None) else (_arrow3981() if (fragment_descriptors is None) else _arrow3982())
    def f_1(c_1: Comment, run: Any=run, fs: Any=fs) -> LDNode:
        return BaseTypes_composeComment_Z13201A7E(c_1)

    comments: Array[LDNode] | None = Option_fromSeq(ResizeArray_map(f_1, run.Comments))
    return LDDataset.create_arcrun(run.Identifier, None, run.Title, run.Description, creators, data_files, measurement_method, measurement_technique, variable_measureds, workflow_invocations, workflow_invocations, comments)


def RunConversion_decomposeRun_Z6839B9E8(run: LDNode, graph: LDGraph | None=None, context: LDContext | None=None) -> ArcRun:
    def predicate_1(wi: LDNode, run: Any=run, graph: Any=graph, context: Any=context) -> bool:
        def predicate(i: LDNode, wi: Any=wi) -> bool:
            return LDFile.try_get_example_of_work_as_formal_parameter(i, graph, context) is not None

        return exists(predicate, LDLabProcess.get_objects(wi, graph, context))

    main_workflow_invocation: LDNode = find(predicate_1, LDDataset.get_abouts_as_workflow_invocation(run, graph, context))
    pattern_input: tuple[CWLProcessingUnit, Array[CWLParameterReference]] = RunConversion_decomposeMainWorkflowInvocation_Z1346FE9D(main_workflow_invocation, LDDataset.get_identifier_as_string(run, context), context, graph)
    def mapping(m: LDNode, run: Any=run, graph: Any=graph, context: Any=context) -> OntologyAnnotation:
        return BaseTypes_decomposeDefinedTerm_Z2F770004(m, context)

    measurement_method: OntologyAnnotation | None = map(mapping, LDDataset.try_get_measurement_method_as_defined_term(run, graph, context))
    def mapping_1(m_1: LDNode, run: Any=run, graph: Any=graph, context: Any=context) -> OntologyAnnotation:
        return BaseTypes_decomposeDefinedTerm_Z2F770004(m_1, context)

    measurement_technique: OntologyAnnotation | None = map(mapping_1, LDDataset.try_get_measurement_technique_as_defined_term(run, graph, context))
    def mapping_2(v: LDNode, run: Any=run, graph: Any=graph, context: Any=context) -> OntologyAnnotation:
        return BaseTypes_decomposePropertyValueToOA_Z2F770004(v, context)

    variable_measured: OntologyAnnotation | None = map(mapping_2, LDDataset.try_get_variable_measured_as_measurement_type(run, graph, context))
    def f(c: LDNode, run: Any=run, graph: Any=graph, context: Any=context) -> Person:
        return PersonConversion_decomposePerson_Z6839B9E8(c, graph, context)

    contacts: Array[Person] = ResizeArray_map(f, LDDataset.get_creators(run, graph, context))
    def f_1(c_1: LDNode, run: Any=run, graph: Any=graph, context: Any=context) -> Comment:
        return BaseTypes_decomposeComment_Z2F770004(c_1, context)

    comments: Array[Comment] = ResizeArray_map(f_1, LDDataset.get_comments(run, graph, context))
    data_map: DataMap | None
    v_1: DataMap = DatamapConversion_decomposeFragmentDescriptors_Z6E59645F(LDDataset.get_variable_measured_as_fragment_descriptors(run, graph, context), graph, context)
    data_map = Option_fromValueWithDefault(DataMap.init(), v_1)
    def f_2(wi_1: LDNode, run: Any=run, graph: Any=graph, context: Any=context) -> bool:
        return wi_1.Id != main_workflow_invocation.Id

    tables: ArcTables = ARCtrl_ArcTables__ArcTables_fromProcesses_Static_Z27F0B586(of_seq(ResizeArray_filter(f_2, LDDataset.get_abouts_as_lab_process(run, graph, context))), graph, context)
    return ArcRun.create(LDDataset.get_identifier_as_string(run, context), LDDataset.try_get_name_as_string(run, context), LDDataset.try_get_description_as_string(run, context), variable_measured, measurement_method, measurement_technique, None, tables.Tables, data_map, contacts, pattern_input[0], pattern_input[1], comments)


__all__ = ["RunConversion_reflection", "RunConversion_composeCWLInputFilePath_Z384F8060", "RunConversion_decomposeCWLInputFilePath_Z384F8060", "RunConversion_composeCWLInputValue_70DD9184", "RunConversion_decomposeCWLInputValue_Z1346FE9D", "RunConversion_composeWorkflowInvocationFromArcRun_Z8CC08AC", "RunConversion_decomposeMainWorkflowInvocation_Z1346FE9D", "RunConversion_composeRun_Z8CC08AC", "RunConversion_decomposeRun_Z6839B9E8"]




from __future__ import annotations
from typing import (Any, Generic, TypeVar)
from ...fable_library.reflection import (TypeInfo, class_type)

_T = TypeVar("_T")

def _expr270(gen0: TypeInfo) -> TypeInfo:
    return class_type("FsSpreadsheet.DSL.Expression.OptionalSource`1", [gen0], OptionalSource_1)


class OptionalSource_1(Generic[_T]):
    def __init__(self, s: Any | None=None) -> None:
        self.s: _T = s


OptionalSource_1_reflection = _expr270

def OptionalSource_1__ctor_2B595(s: Any | None=None) -> OptionalSource_1[_T]:
    return OptionalSource_1(s)


def OptionalSource_1__get_Source(this: OptionalSource_1[Any]) -> Any:
    return this.s


def _expr271(gen0: TypeInfo) -> TypeInfo:
    return class_type("FsSpreadsheet.DSL.Expression.RequiredSource`1", [gen0], RequiredSource_1)


class RequiredSource_1(Generic[_T]):
    def __init__(self, s: Any | None=None) -> None:
        self.s: _T = s


RequiredSource_1_reflection = _expr271

def RequiredSource_1__ctor_2B595(s: Any | None=None) -> RequiredSource_1[_T]:
    return RequiredSource_1(s)


def RequiredSource_1__get_Source(this: RequiredSource_1[Any]) -> Any:
    return this.s


def _expr272(gen0: TypeInfo) -> TypeInfo:
    return class_type("FsSpreadsheet.DSL.Expression.ExpressionSource`1", [gen0], ExpressionSource_1)


class ExpressionSource_1(Generic[_T]):
    def __init__(self, s: Any | None=None) -> None:
        self.s: _T = s


ExpressionSource_1_reflection = _expr272

def ExpressionSource_1__ctor_2B595(s: Any | None=None) -> ExpressionSource_1[_T]:
    return ExpressionSource_1(s)


def ExpressionSource_1__get_Source(this: ExpressionSource_1[Any]) -> Any:
    return this.s


__all__ = ["OptionalSource_1_reflection", "OptionalSource_1__get_Source", "RequiredSource_1_reflection", "RequiredSource_1__get_Source", "ExpressionSource_1_reflection", "ExpressionSource_1__get_Source"]




from __future__ import annotations
from collections.abc import Callable
from typing import Any
from ...fable_library.list import (choose, FSharpList, filter, concat, collect, empty, map, zip, singleton)
from ...fable_library.map import (of_list as of_list_1, FSharpMap__TryFind)
from ...fable_library.option import (bind, default_arg, value as value_4)
from ...fable_library.seq2 import (List_distinct, List_groupBy)
from ...fable_library.set import (of_list, FSharpSet__Contains)
from ...fable_library.util import (string_hash, equals, safe_hash, compare_primitives, equal_arrays, array_hash)
from ..data import Data
from ..ontology_annotation import OntologyAnnotation
from .factor import Factor
from .factor_value import FactorValue
from .material import Material
from .material_attribute import MaterialAttribute
from .material_attribute_value import MaterialAttributeValue
from .process import (Process, Process_tryGetInputsWithParameterBy, Process_tryGetOutputsWithParameterBy, Process_getParameters_763471FF, Process_getCharacteristics_763471FF, Process_tryGetInputsWithCharacteristicBy, Process_tryGetOutputsWithCharacteristicBy, Process_tryGetOutputsWithFactorBy, Process_getFactors_763471FF, Process_getUnits_763471FF, Process_getData_763471FF, Process_getSources_763471FF, Process_getSamples_763471FF, Process_getMaterials_763471FF)
from .process_input import (ProcessInput, ProcessInput_getName_5B3D5BA9, ProcessInput__get_Name)
from .process_output import (ProcessOutput, ProcessOutput_getName_Z42C11600, ProcessOutput__get_Name)
from .process_parameter_value import ProcessParameterValue
from .protocol import Protocol
from .protocol_parameter import ProtocolParameter
from .sample import Sample
from .source import Source

def get_protocol_names(process_sequence: FSharpList[Process]) -> FSharpList[str]:
    def chooser(p: Process, process_sequence: Any=process_sequence) -> str | None:
        def binder(protocol: Protocol, p: Any=p) -> str | None:
            return protocol.Name

        return bind(binder, p.ExecutesProtocol)

    class ObjectExpr786:
        @property
        def Equals(self) -> Callable[[str, str], bool]:
            def _arrow785(x: str, y: str) -> bool:
                return x == y

            return _arrow785

        @property
        def GetHashCode(self) -> Callable[[str], int]:
            return string_hash

    return List_distinct(choose(chooser, process_sequence), ObjectExpr786())


def get_protocols(process_sequence: FSharpList[Process]) -> FSharpList[Protocol]:
    def chooser(p: Process, process_sequence: Any=process_sequence) -> Protocol | None:
        return p.ExecutesProtocol

    class ObjectExpr787:
        @property
        def Equals(self) -> Callable[[Protocol, Protocol], bool]:
            return equals

        @property
        def GetHashCode(self) -> Callable[[Protocol], int]:
            return safe_hash

    return List_distinct(choose(chooser, process_sequence), ObjectExpr787())


def filter_by_protocol_by(predicate: Callable[[Protocol], bool], process_sequence: FSharpList[Process]) -> FSharpList[Process]:
    def predicate_1(p: Process, predicate: Any=predicate, process_sequence: Any=process_sequence) -> bool:
        match_value: Protocol | None = p.ExecutesProtocol
        (pattern_matching_result,) = (None,)
        if match_value is not None:
            if predicate(match_value):
                pattern_matching_result = 0

            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1

        if pattern_matching_result == 0:
            return True

        elif pattern_matching_result == 1:
            return False


    return filter(predicate_1, process_sequence)


def filter_by_protocol_name(protocol_name: str, process_sequence: FSharpList[Process]) -> FSharpList[Process]:
    def _arrow788(p: Protocol, protocol_name: Any=protocol_name, process_sequence: Any=process_sequence) -> bool:
        return equals(p.Name, protocol_name)

    return filter_by_protocol_by(_arrow788, process_sequence)


def get_inputs_with_parameter_by(predicate: Callable[[ProtocolParameter], bool], process_sequence: FSharpList[Process]) -> FSharpList[tuple[ProcessInput, ProcessParameterValue]]:
    def chooser(p: Process, predicate: Any=predicate, process_sequence: Any=process_sequence) -> FSharpList[tuple[ProcessInput, ProcessParameterValue]] | None:
        return Process_tryGetInputsWithParameterBy(predicate, p)

    return concat(choose(chooser, process_sequence))


def get_outputs_with_parameter_by(predicate: Callable[[ProtocolParameter], bool], process_sequence: FSharpList[Process]) -> FSharpList[tuple[ProcessOutput, ProcessParameterValue]]:
    def chooser(p: Process, predicate: Any=predicate, process_sequence: Any=process_sequence) -> FSharpList[tuple[ProcessOutput, ProcessParameterValue]] | None:
        return Process_tryGetOutputsWithParameterBy(predicate, p)

    return concat(choose(chooser, process_sequence))


def get_parameters(process_sequence: FSharpList[Process]) -> FSharpList[ProtocolParameter]:
    def mapping(p: Process, process_sequence: Any=process_sequence) -> FSharpList[ProtocolParameter]:
        return Process_getParameters_763471FF(p)

    class ObjectExpr789:
        @property
        def Equals(self) -> Callable[[ProtocolParameter, ProtocolParameter], bool]:
            return equals

        @property
        def GetHashCode(self) -> Callable[[ProtocolParameter], int]:
            return safe_hash

    return List_distinct(collect(mapping, process_sequence), ObjectExpr789())


def get_characteristics(process_sequence: FSharpList[Process]) -> FSharpList[MaterialAttribute]:
    def mapping(p: Process, process_sequence: Any=process_sequence) -> FSharpList[MaterialAttribute]:
        return Process_getCharacteristics_763471FF(p)

    class ObjectExpr790:
        @property
        def Equals(self) -> Callable[[MaterialAttribute, MaterialAttribute], bool]:
            return equals

        @property
        def GetHashCode(self) -> Callable[[MaterialAttribute], int]:
            return safe_hash

    return List_distinct(collect(mapping, process_sequence), ObjectExpr790())


def get_inputs_with_characteristic_by(predicate: Callable[[MaterialAttribute], bool], process_sequence: FSharpList[Process]) -> FSharpList[tuple[ProcessInput, MaterialAttributeValue]]:
    def chooser(p: Process, predicate: Any=predicate, process_sequence: Any=process_sequence) -> FSharpList[tuple[ProcessInput, MaterialAttributeValue]] | None:
        return Process_tryGetInputsWithCharacteristicBy(predicate, p)

    return concat(choose(chooser, process_sequence))


def get_outputs_with_characteristic_by(predicate: Callable[[MaterialAttribute], bool], process_sequence: FSharpList[Process]) -> FSharpList[tuple[ProcessOutput, MaterialAttributeValue]]:
    def chooser(p: Process, predicate: Any=predicate, process_sequence: Any=process_sequence) -> FSharpList[tuple[ProcessOutput, MaterialAttributeValue]] | None:
        return Process_tryGetOutputsWithCharacteristicBy(predicate, p)

    return concat(choose(chooser, process_sequence))


def get_outputs_with_factor_by(predicate: Callable[[Factor], bool], process_sequence: FSharpList[Process]) -> FSharpList[tuple[ProcessOutput, FactorValue]]:
    def chooser(p: Process, predicate: Any=predicate, process_sequence: Any=process_sequence) -> FSharpList[tuple[ProcessOutput, FactorValue]] | None:
        return Process_tryGetOutputsWithFactorBy(predicate, p)

    return concat(choose(chooser, process_sequence))


def get_factors(process_sequence: FSharpList[Process]) -> FSharpList[Factor]:
    def mapping(p: Process, process_sequence: Any=process_sequence) -> FSharpList[Factor]:
        return Process_getFactors_763471FF(p)

    class ObjectExpr791:
        @property
        def Equals(self) -> Callable[[Factor, Factor], bool]:
            return equals

        @property
        def GetHashCode(self) -> Callable[[Factor], int]:
            return safe_hash

    return List_distinct(collect(mapping, process_sequence), ObjectExpr791())


def get_root_inputs(process_sequence: FSharpList[Process]) -> FSharpList[ProcessInput]:
    def mapping(p: Process, process_sequence: Any=process_sequence) -> FSharpList[ProcessInput]:
        return default_arg(p.Inputs, empty())

    inputs: FSharpList[ProcessInput] = collect(mapping, process_sequence)
    def mapping_2(p_1: Process, process_sequence: Any=process_sequence) -> FSharpList[str]:
        def mapping_1(po: ProcessOutput, p_1: Any=p_1) -> str:
            return ProcessOutput_getName_Z42C11600(po)

        return map(mapping_1, default_arg(p_1.Outputs, empty()))

    class ObjectExpr792:
        @property
        def Compare(self) -> Callable[[str, str], int]:
            return compare_primitives

    outputs: Any = of_list(collect(mapping_2, process_sequence), ObjectExpr792())
    def predicate(i: ProcessInput, process_sequence: Any=process_sequence) -> bool:
        return not FSharpSet__Contains(outputs, ProcessInput_getName_5B3D5BA9(i))

    return filter(predicate, inputs)


def get_final_outputs(process_sequence: FSharpList[Process]) -> FSharpList[ProcessOutput]:
    def mapping_1(p: Process, process_sequence: Any=process_sequence) -> FSharpList[str]:
        def mapping(pi: ProcessInput, p: Any=p) -> str:
            return ProcessInput_getName_5B3D5BA9(pi)

        return map(mapping, default_arg(p.Inputs, empty()))

    class ObjectExpr793:
        @property
        def Compare(self) -> Callable[[str, str], int]:
            return compare_primitives

    inputs: Any = of_list(collect(mapping_1, process_sequence), ObjectExpr793())
    def predicate(o: ProcessOutput, process_sequence: Any=process_sequence) -> bool:
        return not FSharpSet__Contains(inputs, ProcessOutput_getName_Z42C11600(o))

    def mapping_2(p_1: Process, process_sequence: Any=process_sequence) -> FSharpList[ProcessOutput]:
        return default_arg(p_1.Outputs, empty())

    return filter(predicate, collect(mapping_2, process_sequence))


def get_root_input_of(process_sequence: FSharpList[Process], sample: str) -> FSharpList[str]:
    def mapping_4(tupled_arg: tuple[str, FSharpList[tuple[str, str]]], process_sequence: Any=process_sequence, sample: Any=sample) -> tuple[str, FSharpList[str]]:
        def mapping_3(tuple_1: tuple[str, str], tupled_arg: Any=tupled_arg) -> str:
            return tuple_1[1]

        return (tupled_arg[0], map(mapping_3, tupled_arg[1]))

    def projection(tuple: tuple[str, str], process_sequence: Any=process_sequence, sample: Any=sample) -> str:
        return tuple[0]

    def mapping_2(p: Process, process_sequence: Any=process_sequence, sample: Any=sample) -> FSharpList[tuple[str, str]]:
        def mapping(o: ProcessOutput, p: Any=p) -> str:
            return ProcessOutput__get_Name(o)

        def mapping_1(i: ProcessInput, p: Any=p) -> str:
            return ProcessInput__get_Name(i)

        class ObjectExpr794:
            @property
            def Equals(self) -> Callable[[tuple[str, str], tuple[str, str]], bool]:
                return equal_arrays

            @property
            def GetHashCode(self) -> Callable[[tuple[str, str]], int]:
                return array_hash

        return List_distinct(zip(map(mapping, value_4(p.Outputs)), map(mapping_1, value_4(p.Inputs))), ObjectExpr794())

    class ObjectExpr796:
        @property
        def Equals(self) -> Callable[[str, str], bool]:
            def _arrow795(x_1: str, y_1: str) -> bool:
                return x_1 == y_1

            return _arrow795

        @property
        def GetHashCode(self) -> Callable[[str], int]:
            return string_hash

    class ObjectExpr797:
        @property
        def Compare(self) -> Callable[[str, str], int]:
            return compare_primitives

    mappings: Any = of_list_1(map(mapping_4, List_groupBy(projection, collect(mapping_2, process_sequence), ObjectExpr796())), ObjectExpr797())
    def loop(last_state_mut: FSharpList[str], state_mut: FSharpList[str], process_sequence: Any=process_sequence, sample: Any=sample) -> FSharpList[str]:
        while True:
            (last_state, state) = (last_state_mut, state_mut)
            if equals(last_state, state):
                return state

            else: 
                last_state_mut = state
                def mapping_5(s: str, last_state: Any=last_state, state: Any=state) -> FSharpList[str]:
                    return default_arg(FSharpMap__TryFind(mappings, s), singleton(s))

                state_mut = collect(mapping_5, state)
                continue

            break

    return loop(empty(), singleton(sample))


def get_final_outputs_of(process_sequence: FSharpList[Process], sample: str) -> FSharpList[str]:
    def mapping_4(tupled_arg: tuple[str, FSharpList[tuple[str, str]]], process_sequence: Any=process_sequence, sample: Any=sample) -> tuple[str, FSharpList[str]]:
        def mapping_3(tuple_1: tuple[str, str], tupled_arg: Any=tupled_arg) -> str:
            return tuple_1[1]

        return (tupled_arg[0], map(mapping_3, tupled_arg[1]))

    def projection(tuple: tuple[str, str], process_sequence: Any=process_sequence, sample: Any=sample) -> str:
        return tuple[0]

    def mapping_2(p: Process, process_sequence: Any=process_sequence, sample: Any=sample) -> FSharpList[tuple[str, str]]:
        def mapping(i: ProcessInput, p: Any=p) -> str:
            return ProcessInput__get_Name(i)

        def mapping_1(o: ProcessOutput, p: Any=p) -> str:
            return ProcessOutput__get_Name(o)

        class ObjectExpr798:
            @property
            def Equals(self) -> Callable[[tuple[str, str], tuple[str, str]], bool]:
                return equal_arrays

            @property
            def GetHashCode(self) -> Callable[[tuple[str, str]], int]:
                return array_hash

        return List_distinct(zip(map(mapping, value_4(p.Inputs)), map(mapping_1, value_4(p.Outputs))), ObjectExpr798())

    class ObjectExpr800:
        @property
        def Equals(self) -> Callable[[str, str], bool]:
            def _arrow799(x_1: str, y_1: str) -> bool:
                return x_1 == y_1

            return _arrow799

        @property
        def GetHashCode(self) -> Callable[[str], int]:
            return string_hash

    class ObjectExpr801:
        @property
        def Compare(self) -> Callable[[str, str], int]:
            return compare_primitives

    mappings: Any = of_list_1(map(mapping_4, List_groupBy(projection, collect(mapping_2, process_sequence), ObjectExpr800())), ObjectExpr801())
    def loop(last_state_mut: FSharpList[str], state_mut: FSharpList[str], process_sequence: Any=process_sequence, sample: Any=sample) -> FSharpList[str]:
        while True:
            (last_state, state) = (last_state_mut, state_mut)
            if equals(last_state, state):
                return state

            else: 
                last_state_mut = state
                def mapping_5(s: str, last_state: Any=last_state, state: Any=state) -> FSharpList[str]:
                    return default_arg(FSharpMap__TryFind(mappings, s), singleton(s))

                state_mut = collect(mapping_5, state)
                continue

            break

    return loop(empty(), singleton(sample))


def get_units(process_sequence: FSharpList[Process]) -> FSharpList[OntologyAnnotation]:
    def _arrow802(p: Process, process_sequence: Any=process_sequence) -> FSharpList[OntologyAnnotation]:
        return Process_getUnits_763471FF(p)

    class ObjectExpr803:
        @property
        def Equals(self) -> Callable[[OntologyAnnotation, OntologyAnnotation], bool]:
            return equals

        @property
        def GetHashCode(self) -> Callable[[OntologyAnnotation], int]:
            return safe_hash

    return List_distinct(collect(_arrow802, process_sequence), ObjectExpr803())


def get_data(process_sequence: FSharpList[Process]) -> FSharpList[Data]:
    def mapping(p: Process, process_sequence: Any=process_sequence) -> FSharpList[Data]:
        return Process_getData_763471FF(p)

    class ObjectExpr804:
        @property
        def Equals(self) -> Callable[[Data, Data], bool]:
            return equals

        @property
        def GetHashCode(self) -> Callable[[Data], int]:
            return safe_hash

    return List_distinct(collect(mapping, process_sequence), ObjectExpr804())


def get_sources(process_sequence: FSharpList[Process]) -> FSharpList[Source]:
    def mapping(p: Process, process_sequence: Any=process_sequence) -> FSharpList[Source]:
        return Process_getSources_763471FF(p)

    class ObjectExpr805:
        @property
        def Equals(self) -> Callable[[Source, Source], bool]:
            return equals

        @property
        def GetHashCode(self) -> Callable[[Source], int]:
            return safe_hash

    return List_distinct(collect(mapping, process_sequence), ObjectExpr805())


def get_samples(process_sequence: FSharpList[Process]) -> FSharpList[Sample]:
    def mapping(p: Process, process_sequence: Any=process_sequence) -> FSharpList[Sample]:
        return Process_getSamples_763471FF(p)

    class ObjectExpr806:
        @property
        def Equals(self) -> Callable[[Sample, Sample], bool]:
            return equals

        @property
        def GetHashCode(self) -> Callable[[Sample], int]:
            return safe_hash

    return List_distinct(collect(mapping, process_sequence), ObjectExpr806())


def get_materials(process_sequence: FSharpList[Process]) -> FSharpList[Material]:
    def mapping(p: Process, process_sequence: Any=process_sequence) -> FSharpList[Material]:
        return Process_getMaterials_763471FF(p)

    class ObjectExpr807:
        @property
        def Equals(self) -> Callable[[Material, Material], bool]:
            return equals

        @property
        def GetHashCode(self) -> Callable[[Material], int]:
            return safe_hash

    return List_distinct(collect(mapping, process_sequence), ObjectExpr807())


__all__ = ["get_protocol_names", "get_protocols", "filter_by_protocol_by", "filter_by_protocol_name", "get_inputs_with_parameter_by", "get_outputs_with_parameter_by", "get_parameters", "get_characteristics", "get_inputs_with_characteristic_by", "get_outputs_with_characteristic_by", "get_outputs_with_factor_by", "get_factors", "get_root_inputs", "get_final_outputs", "get_root_input_of", "get_final_outputs_of", "get_units", "get_data", "get_sources", "get_samples", "get_materials"]




from typing import Any
from ...fable_library.reg_exp import (create, match)
from .regex import ActivePatterns__007CRegex_007C__007C

orcid_regex: Any = create("[0-9]{4}-[0-9]{4}-[0-9]{4}-[0-9]{3}[0-9X]")

def try_get_orcid_number(orcid: str) -> str | None:
    m: Any = match(orcid_regex, orcid)
    if m is not None:
        return m[0]

    else: 
        return None



orcid_prefix: str = "http://orcid.org/"

def _007CORCID_007C__007C(input: str) -> str | None:
    active_pattern_result: Any | None = ActivePatterns__007CRegex_007C__007C("[0-9]{4}-[0-9]{4}-[0-9]{4}-[0-9]{3}[0-9X]", input)
    if active_pattern_result is not None:
        r: Any = active_pattern_result
        return r[0]

    else: 
        return None



def try_get_orcid_url(orcid: str) -> str | None:
    active_pattern_result: str | None = _007CORCID_007C__007C(orcid)
    if active_pattern_result is not None:
        orcid_1: str = active_pattern_result
        return ((("" + orcid_prefix) + "") + orcid_1) + ""

    else: 
        return None



__all__ = ["orcid_regex", "try_get_orcid_number", "orcid_prefix", "_007CORCID_007C__007C", "try_get_orcid_url"]




from __future__ import annotations
from collections.abc import Callable
from typing import (Any, TypeVar)
from ..arctrl_core.comment import Comment
from ..arctrl_core.conversion import (Person_setCommentFromORCID, Person_setOrcidFromComments)
from ..arctrl_core.ontology_annotation import OntologyAnnotation
from ..arctrl_core.person import Person
from ..fable_library.array_ import map as map_2
from ..fable_library.list import (choose, of_array, FSharpList)
from ..fable_library.option import (map, default_arg)
from ..fable_library.seq import (map as map_1, try_pick)
from ..fable_library.string_ import (replace, split, join)
from ..fable_library.types import Array
from ..fable_library.util import IEnumerable_1
from ..thoth_json_core.decode import (object, IOptionalGetter, string, resize_array, IGetters, IRequiredGetter, map as map_3, array as array_3)
from ..thoth_json_core.types import (IEncodable, IEncoderHelpers_1, Decoder_1)
from .comment import (encoder as encoder_1, decoder as decoder_1, ROCrate_encoderDisambiguatingDescription, ROCrate_decoderDisambiguatingDescription)
from .context.rocrate.isa_organization_context import context_jsonvalue
from .context.rocrate.isa_person_context import (context_jsonvalue as context_jsonvalue_1, context_minimal_json_value)
from .decode import Decode_objectNoAdditionalProperties
from .encode import (try_include, try_include_seq)
from .idtable import encode
from .ontology_annotation import (OntologyAnnotation_encoder, OntologyAnnotation_decoder, OntologyAnnotation_ROCrate_encoderDefinedTerm, OntologyAnnotation_ROCrate_decoderDefinedTerm)

__A_ = TypeVar("__A_")

def encoder(person: Person) -> IEncodable:
    def chooser(tupled_arg: tuple[str, IEncodable | None], person: Any=person) -> tuple[str, IEncodable] | None:
        def mapping(v_1: IEncodable, tupled_arg: Any=tupled_arg) -> tuple[str, IEncodable]:
            return (tupled_arg[0], v_1)

        return map(mapping, tupled_arg[1])

    def _arrow2294(value: str, person: Any=person) -> IEncodable:
        class ObjectExpr2293(IEncodable):
            def Encode(self, helpers: IEncoderHelpers_1[Any]) -> Any:
                return helpers.encode_string(value)

        return ObjectExpr2293()

    def _arrow2296(value_2: str, person: Any=person) -> IEncodable:
        class ObjectExpr2295(IEncodable):
            def Encode(self, helpers_1: IEncoderHelpers_1[Any]) -> Any:
                return helpers_1.encode_string(value_2)

        return ObjectExpr2295()

    def _arrow2299(value_4: str, person: Any=person) -> IEncodable:
        class ObjectExpr2298(IEncodable):
            def Encode(self, helpers_2: IEncoderHelpers_1[Any]) -> Any:
                return helpers_2.encode_string(value_4)

        return ObjectExpr2298()

    def _arrow2301(value_6: str, person: Any=person) -> IEncodable:
        class ObjectExpr2300(IEncodable):
            def Encode(self, helpers_3: IEncoderHelpers_1[Any]) -> Any:
                return helpers_3.encode_string(value_6)

        return ObjectExpr2300()

    def _arrow2303(value_8: str, person: Any=person) -> IEncodable:
        class ObjectExpr2302(IEncodable):
            def Encode(self, helpers_4: IEncoderHelpers_1[Any]) -> Any:
                return helpers_4.encode_string(value_8)

        return ObjectExpr2302()

    def _arrow2305(value_10: str, person: Any=person) -> IEncodable:
        class ObjectExpr2304(IEncodable):
            def Encode(self, helpers_5: IEncoderHelpers_1[Any]) -> Any:
                return helpers_5.encode_string(value_10)

        return ObjectExpr2304()

    def _arrow2307(value_12: str, person: Any=person) -> IEncodable:
        class ObjectExpr2306(IEncodable):
            def Encode(self, helpers_6: IEncoderHelpers_1[Any]) -> Any:
                return helpers_6.encode_string(value_12)

        return ObjectExpr2306()

    def _arrow2309(value_14: str, person: Any=person) -> IEncodable:
        class ObjectExpr2308(IEncodable):
            def Encode(self, helpers_7: IEncoderHelpers_1[Any]) -> Any:
                return helpers_7.encode_string(value_14)

        return ObjectExpr2308()

    def _arrow2311(value_16: str, person: Any=person) -> IEncodable:
        class ObjectExpr2310(IEncodable):
            def Encode(self, helpers_8: IEncoderHelpers_1[Any]) -> Any:
                return helpers_8.encode_string(value_16)

        return ObjectExpr2310()

    def _arrow2312(oa: OntologyAnnotation, person: Any=person) -> IEncodable:
        return OntologyAnnotation_encoder(oa)

    def _arrow2313(comment: Comment, person: Any=person) -> IEncodable:
        return encoder_1(comment)

    values: FSharpList[tuple[str, IEncodable]] = choose(chooser, of_array([try_include("firstName", _arrow2294, person.FirstName), try_include("lastName", _arrow2296, person.LastName), try_include("midInitials", _arrow2299, person.MidInitials), try_include("orcid", _arrow2301, person.ORCID), try_include("email", _arrow2303, person.EMail), try_include("phone", _arrow2305, person.Phone), try_include("fax", _arrow2307, person.Fax), try_include("address", _arrow2309, person.Address), try_include("affiliation", _arrow2311, person.Affiliation), try_include_seq("roles", _arrow2312, person.Roles), try_include_seq("comments", _arrow2313, person.Comments)]))
    class ObjectExpr2314(IEncodable):
        def Encode(self, helpers_9: IEncoderHelpers_1[Any], person: Any=person) -> Any:
            def mapping_1(tupled_arg_1: tuple[str, IEncodable]) -> tuple[str, __A_]:
                return (tupled_arg_1[0], tupled_arg_1[1].Encode(helpers_9))

            arg: IEnumerable_1[tuple[str, __A_]] = map_1(mapping_1, values)
            return helpers_9.encode_object(arg)

    return ObjectExpr2314()


def _arrow2327(get: IGetters) -> Person:
    def _arrow2315(__unit: None=None) -> str | None:
        object_arg: IOptionalGetter = get.Optional
        return object_arg.Field("orcid", string)

    def _arrow2316(__unit: None=None) -> str | None:
        object_arg_1: IOptionalGetter = get.Optional
        return object_arg_1.Field("lastName", string)

    def _arrow2317(__unit: None=None) -> str | None:
        object_arg_2: IOptionalGetter = get.Optional
        return object_arg_2.Field("firstName", string)

    def _arrow2318(__unit: None=None) -> str | None:
        object_arg_3: IOptionalGetter = get.Optional
        return object_arg_3.Field("midInitials", string)

    def _arrow2319(__unit: None=None) -> str | None:
        object_arg_4: IOptionalGetter = get.Optional
        return object_arg_4.Field("email", string)

    def _arrow2320(__unit: None=None) -> str | None:
        object_arg_5: IOptionalGetter = get.Optional
        return object_arg_5.Field("phone", string)

    def _arrow2321(__unit: None=None) -> str | None:
        object_arg_6: IOptionalGetter = get.Optional
        return object_arg_6.Field("fax", string)

    def _arrow2322(__unit: None=None) -> str | None:
        object_arg_7: IOptionalGetter = get.Optional
        return object_arg_7.Field("address", string)

    def _arrow2323(__unit: None=None) -> str | None:
        object_arg_8: IOptionalGetter = get.Optional
        return object_arg_8.Field("affiliation", string)

    def _arrow2324(__unit: None=None) -> Array[OntologyAnnotation] | None:
        arg_19: Decoder_1[Array[OntologyAnnotation]] = resize_array(OntologyAnnotation_decoder)
        object_arg_9: IOptionalGetter = get.Optional
        return object_arg_9.Field("roles", arg_19)

    def _arrow2325(__unit: None=None) -> Array[Comment] | None:
        arg_21: Decoder_1[Array[Comment]] = resize_array(decoder_1)
        object_arg_10: IOptionalGetter = get.Optional
        return object_arg_10.Field("comments", arg_21)

    return Person(_arrow2315(), _arrow2316(), _arrow2317(), _arrow2318(), _arrow2319(), _arrow2320(), _arrow2321(), _arrow2322(), _arrow2323(), _arrow2324(), _arrow2325())


decoder: Decoder_1[Person] = object(_arrow2327)

def ROCrate_genID(p: Person) -> str:
    def chooser(c: Comment, p: Any=p) -> str | None:
        match_value: str | None = c.Name
        match_value_1: str | None = c.Value
        (pattern_matching_result, n, v) = (None, None, None)
        if match_value is not None:
            if match_value_1 is not None:
                pattern_matching_result = 0
                n = match_value
                v = match_value_1

            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1

        if pattern_matching_result == 0:
            if True if (True if (n == "orcid") else (n == "Orcid")) else (n == "ORCID"):
                return v

            else: 
                return None


        elif pattern_matching_result == 1:
            return None


    orcid: str | None = try_pick(chooser, p.Comments)
    if orcid is None:
        match_value_1: str | None = p.EMail
        if match_value_1 is None:
            match_value_2: str | None = p.FirstName
            match_value_3: str | None = p.MidInitials
            match_value_4: str | None = p.LastName
            (pattern_matching_result_1, fn, ln, mn, fn_1, ln_1, ln_2, fn_2) = (None, None, None, None, None, None, None, None)
            if match_value_2 is None:
                if match_value_3 is None:
                    if match_value_4 is not None:
                        pattern_matching_result_1 = 2
                        ln_2 = match_value_4

                    else: 
                        pattern_matching_result_1 = 4


                else: 
                    pattern_matching_result_1 = 4


            elif match_value_3 is None:
                if match_value_4 is None:
                    pattern_matching_result_1 = 3
                    fn_2 = match_value_2

                else: 
                    pattern_matching_result_1 = 1
                    fn_1 = match_value_2
                    ln_1 = match_value_4


            elif match_value_4 is not None:
                pattern_matching_result_1 = 0
                fn = match_value_2
                ln = match_value_4
                mn = match_value_3

            else: 
                pattern_matching_result_1 = 4

            if pattern_matching_result_1 == 0:
                return (((("#" + replace(fn, " ", "_")) + "_") + replace(mn, " ", "_")) + "_") + replace(ln, " ", "_")

            elif pattern_matching_result_1 == 1:
                return (("#" + replace(fn_1, " ", "_")) + "_") + replace(ln_1, " ", "_")

            elif pattern_matching_result_1 == 2:
                return "#" + replace(ln_2, " ", "_")

            elif pattern_matching_result_1 == 3:
                return "#" + replace(fn_2, " ", "_")

            elif pattern_matching_result_1 == 4:
                return "#EmptyPerson"


        else: 
            return match_value_1


    else: 
        return orcid



def ROCrate_Affiliation_encoder(affiliation: str) -> IEncodable:
    class ObjectExpr2336(IEncodable):
        def Encode(self, helpers: IEncoderHelpers_1[Any], affiliation: Any=affiliation) -> Any:
            return helpers.encode_string("Organization")

    def _arrow2339(__unit: None=None, affiliation: Any=affiliation) -> IEncodable:
        value_1: str = replace(("#Organization_" + affiliation) + "", " ", "_")
        class ObjectExpr2338(IEncodable):
            def Encode(self, helpers_1: IEncoderHelpers_1[Any]) -> Any:
                return helpers_1.encode_string(value_1)

        return ObjectExpr2338()

    class ObjectExpr2341(IEncodable):
        def Encode(self, helpers_2: IEncoderHelpers_1[Any], affiliation: Any=affiliation) -> Any:
            return helpers_2.encode_string(affiliation)

    values: FSharpList[tuple[str, IEncodable]] = of_array([("@type", ObjectExpr2336()), ("@id", _arrow2339()), ("name", ObjectExpr2341()), ("@context", context_jsonvalue)])
    class ObjectExpr2344(IEncodable):
        def Encode(self, helpers_3: IEncoderHelpers_1[Any], affiliation: Any=affiliation) -> Any:
            def mapping(tupled_arg: tuple[str, IEncodable]) -> tuple[str, __A_]:
                return (tupled_arg[0], tupled_arg[1].Encode(helpers_3))

            arg: IEnumerable_1[tuple[str, __A_]] = map_1(mapping, values)
            return helpers_3.encode_object(arg)

    return ObjectExpr2344()


def _arrow2347(get: IGetters) -> str:
    object_arg: IRequiredGetter = get.Required
    return object_arg.Field("name", string)


ROCrate_Affiliation_decoder: Decoder_1[str] = object(_arrow2347)

def ROCrate_encoder(oa: Person) -> IEncodable:
    def chooser(tupled_arg: tuple[str, IEncodable | None], oa: Any=oa) -> tuple[str, IEncodable] | None:
        def mapping(v_1: IEncodable, tupled_arg: Any=tupled_arg) -> tuple[str, IEncodable]:
            return (tupled_arg[0], v_1)

        return map(mapping, tupled_arg[1])

    def _arrow2351(__unit: None=None, oa: Any=oa) -> IEncodable:
        value: str = ROCrate_genID(oa)
        class ObjectExpr2350(IEncodable):
            def Encode(self, helpers: IEncoderHelpers_1[Any]) -> Any:
                return helpers.encode_string(value)

        return ObjectExpr2350()

    class ObjectExpr2352(IEncodable):
        def Encode(self, helpers_1: IEncoderHelpers_1[Any], oa: Any=oa) -> Any:
            return helpers_1.encode_string("Person")

    def _arrow2354(value_2: str, oa: Any=oa) -> IEncodable:
        class ObjectExpr2353(IEncodable):
            def Encode(self, helpers_2: IEncoderHelpers_1[Any]) -> Any:
                return helpers_2.encode_string(value_2)

        return ObjectExpr2353()

    def _arrow2356(value_4: str, oa: Any=oa) -> IEncodable:
        class ObjectExpr2355(IEncodable):
            def Encode(self, helpers_3: IEncoderHelpers_1[Any]) -> Any:
                return helpers_3.encode_string(value_4)

        return ObjectExpr2355()

    def _arrow2358(value_6: str, oa: Any=oa) -> IEncodable:
        class ObjectExpr2357(IEncodable):
            def Encode(self, helpers_4: IEncoderHelpers_1[Any]) -> Any:
                return helpers_4.encode_string(value_6)

        return ObjectExpr2357()

    def _arrow2360(value_8: str, oa: Any=oa) -> IEncodable:
        class ObjectExpr2359(IEncodable):
            def Encode(self, helpers_5: IEncoderHelpers_1[Any]) -> Any:
                return helpers_5.encode_string(value_8)

        return ObjectExpr2359()

    def _arrow2362(value_10: str, oa: Any=oa) -> IEncodable:
        class ObjectExpr2361(IEncodable):
            def Encode(self, helpers_6: IEncoderHelpers_1[Any]) -> Any:
                return helpers_6.encode_string(value_10)

        return ObjectExpr2361()

    def _arrow2364(value_12: str, oa: Any=oa) -> IEncodable:
        class ObjectExpr2363(IEncodable):
            def Encode(self, helpers_7: IEncoderHelpers_1[Any]) -> Any:
                return helpers_7.encode_string(value_12)

        return ObjectExpr2363()

    def _arrow2366(value_14: str, oa: Any=oa) -> IEncodable:
        class ObjectExpr2365(IEncodable):
            def Encode(self, helpers_8: IEncoderHelpers_1[Any]) -> Any:
                return helpers_8.encode_string(value_14)

        return ObjectExpr2365()

    def _arrow2368(value_16: str, oa: Any=oa) -> IEncodable:
        class ObjectExpr2367(IEncodable):
            def Encode(self, helpers_9: IEncoderHelpers_1[Any]) -> Any:
                return helpers_9.encode_string(value_16)

        return ObjectExpr2367()

    def _arrow2369(affiliation: str, oa: Any=oa) -> IEncodable:
        return ROCrate_Affiliation_encoder(affiliation)

    def _arrow2370(oa_1: OntologyAnnotation, oa: Any=oa) -> IEncodable:
        return OntologyAnnotation_ROCrate_encoderDefinedTerm(oa_1)

    def _arrow2371(comment: Comment, oa: Any=oa) -> IEncodable:
        return ROCrate_encoderDisambiguatingDescription(comment)

    values: FSharpList[tuple[str, IEncodable]] = choose(chooser, of_array([("@id", _arrow2351()), ("@type", ObjectExpr2352()), try_include("orcid", _arrow2354, oa.ORCID), try_include("firstName", _arrow2356, oa.FirstName), try_include("lastName", _arrow2358, oa.LastName), try_include("midInitials", _arrow2360, oa.MidInitials), try_include("email", _arrow2362, oa.EMail), try_include("phone", _arrow2364, oa.Phone), try_include("fax", _arrow2366, oa.Fax), try_include("address", _arrow2368, oa.Address), try_include("affiliation", _arrow2369, oa.Affiliation), try_include_seq("roles", _arrow2370, oa.Roles), try_include_seq("comments", _arrow2371, oa.Comments), ("@context", context_jsonvalue_1)]))
    class ObjectExpr2372(IEncodable):
        def Encode(self, helpers_10: IEncoderHelpers_1[Any], oa: Any=oa) -> Any:
            def mapping_1(tupled_arg_1: tuple[str, IEncodable]) -> tuple[str, __A_]:
                return (tupled_arg_1[0], tupled_arg_1[1].Encode(helpers_10))

            arg: IEnumerable_1[tuple[str, __A_]] = map_1(mapping_1, values)
            return helpers_10.encode_object(arg)

    return ObjectExpr2372()


def _arrow2384(get: IGetters) -> Person:
    def _arrow2373(__unit: None=None) -> str | None:
        object_arg: IOptionalGetter = get.Optional
        return object_arg.Field("orcid", string)

    def _arrow2374(__unit: None=None) -> str | None:
        object_arg_1: IOptionalGetter = get.Optional
        return object_arg_1.Field("lastName", string)

    def _arrow2375(__unit: None=None) -> str | None:
        object_arg_2: IOptionalGetter = get.Optional
        return object_arg_2.Field("firstName", string)

    def _arrow2376(__unit: None=None) -> str | None:
        object_arg_3: IOptionalGetter = get.Optional
        return object_arg_3.Field("midInitials", string)

    def _arrow2377(__unit: None=None) -> str | None:
        object_arg_4: IOptionalGetter = get.Optional
        return object_arg_4.Field("email", string)

    def _arrow2378(__unit: None=None) -> str | None:
        object_arg_5: IOptionalGetter = get.Optional
        return object_arg_5.Field("phone", string)

    def _arrow2379(__unit: None=None) -> str | None:
        object_arg_6: IOptionalGetter = get.Optional
        return object_arg_6.Field("fax", string)

    def _arrow2380(__unit: None=None) -> str | None:
        object_arg_7: IOptionalGetter = get.Optional
        return object_arg_7.Field("address", string)

    def _arrow2381(__unit: None=None) -> str | None:
        object_arg_8: IOptionalGetter = get.Optional
        return object_arg_8.Field("affiliation", ROCrate_Affiliation_decoder)

    def _arrow2382(__unit: None=None) -> Array[OntologyAnnotation] | None:
        arg_19: Decoder_1[Array[OntologyAnnotation]] = resize_array(OntologyAnnotation_ROCrate_decoderDefinedTerm)
        object_arg_9: IOptionalGetter = get.Optional
        return object_arg_9.Field("roles", arg_19)

    def _arrow2383(__unit: None=None) -> Array[Comment] | None:
        arg_21: Decoder_1[Array[Comment]] = resize_array(ROCrate_decoderDisambiguatingDescription)
        object_arg_10: IOptionalGetter = get.Optional
        return object_arg_10.Field("comments", arg_21)

    return Person(_arrow2373(), _arrow2374(), _arrow2375(), _arrow2376(), _arrow2377(), _arrow2378(), _arrow2379(), _arrow2380(), _arrow2381(), _arrow2382(), _arrow2383())


ROCrate_decoder: Decoder_1[Person] = object(_arrow2384)

def ROCrate_encodeAuthorListString(author_list: str) -> IEncodable:
    def encode_single(name: str, author_list: Any=author_list) -> IEncodable:
        def chooser(tupled_arg: tuple[str, IEncodable | None], name: Any=name) -> tuple[str, IEncodable] | None:
            def mapping_1(v_1: IEncodable, tupled_arg: Any=tupled_arg) -> tuple[str, IEncodable]:
                return (tupled_arg[0], v_1)

            return map(mapping_1, tupled_arg[1])

        class ObjectExpr2386(IEncodable):
            def Encode(self, helpers: IEncoderHelpers_1[Any], name: Any=name) -> Any:
                return helpers.encode_string("Person")

        def _arrow2388(value_1: str, name: Any=name) -> IEncodable:
            class ObjectExpr2387(IEncodable):
                def Encode(self, helpers_1: IEncoderHelpers_1[Any]) -> Any:
                    return helpers_1.encode_string(value_1)

            return ObjectExpr2387()

        values: FSharpList[tuple[str, IEncodable]] = choose(chooser, of_array([("@type", ObjectExpr2386()), try_include("name", _arrow2388, name), ("@context", context_minimal_json_value)]))
        class ObjectExpr2389(IEncodable):
            def Encode(self, helpers_2: IEncoderHelpers_1[Any], name: Any=name) -> Any:
                def mapping_2(tupled_arg_1: tuple[str, IEncodable]) -> tuple[str, __A_]:
                    return (tupled_arg_1[0], tupled_arg_1[1].Encode(helpers_2))

                arg: IEnumerable_1[tuple[str, __A_]] = map_1(mapping_2, values)
                return helpers_2.encode_object(arg)

        return ObjectExpr2389()

    def mapping(s: str, author_list: Any=author_list) -> str:
        return s.strip()

    values_2: Array[IEncodable] = map_2(encode_single, map_2(mapping, split(author_list, ["\t" if (author_list.find("\t") >= 0) else (";" if (author_list.find(";") >= 0) else ",")], None, 0), None), None)
    class ObjectExpr2390(IEncodable):
        def Encode(self, helpers_3: IEncoderHelpers_1[Any], author_list: Any=author_list) -> Any:
            def mapping_3(v_3: IEncodable) -> __A_:
                return v_3.Encode(helpers_3)

            arg_1: Array[__A_] = map_2(mapping_3, values_2, None)
            return helpers_3.encode_array(arg_1)

    return ObjectExpr2390()


def ctor(v: Array[str]) -> str:
    return join(", ", v)


def _arrow2392(get: IGetters) -> str:
    def _arrow2391(__unit: None=None) -> str | None:
        object_arg: IOptionalGetter = get.Optional
        return object_arg.Field("name", string)

    return default_arg(_arrow2391(), "")


ROCrate_decodeAuthorListString: Decoder_1[str] = map_3(ctor, array_3(object(_arrow2392)))

ISAJson_allowedFields: FSharpList[str] = of_array(["@id", "firstName", "lastName", "midInitials", "email", "phone", "fax", "address", "affiliation", "roles", "comments", "@type", "@context"])

def ISAJson_encoder(id_map: Any | None, person: Person) -> IEncodable:
    def f(person_1: Person, id_map: Any=id_map, person: Any=person) -> IEncodable:
        person_2: Person = Person_setCommentFromORCID(person_1)
        def chooser(tupled_arg: tuple[str, IEncodable | None], person_1: Any=person_1) -> tuple[str, IEncodable] | None:
            def mapping(v_1: IEncodable, tupled_arg: Any=tupled_arg) -> tuple[str, IEncodable]:
                return (tupled_arg[0], v_1)

            return map(mapping, tupled_arg[1])

        def _arrow2396(value: str, person_1: Any=person_1) -> IEncodable:
            class ObjectExpr2395(IEncodable):
                def Encode(self, helpers: IEncoderHelpers_1[Any]) -> Any:
                    return helpers.encode_string(value)

            return ObjectExpr2395()

        def _arrow2398(value_2: str, person_1: Any=person_1) -> IEncodable:
            class ObjectExpr2397(IEncodable):
                def Encode(self, helpers_1: IEncoderHelpers_1[Any]) -> Any:
                    return helpers_1.encode_string(value_2)

            return ObjectExpr2397()

        def _arrow2400(value_4: str, person_1: Any=person_1) -> IEncodable:
            class ObjectExpr2399(IEncodable):
                def Encode(self, helpers_2: IEncoderHelpers_1[Any]) -> Any:
                    return helpers_2.encode_string(value_4)

            return ObjectExpr2399()

        def _arrow2402(value_6: str, person_1: Any=person_1) -> IEncodable:
            class ObjectExpr2401(IEncodable):
                def Encode(self, helpers_3: IEncoderHelpers_1[Any]) -> Any:
                    return helpers_3.encode_string(value_6)

            return ObjectExpr2401()

        def _arrow2404(value_8: str, person_1: Any=person_1) -> IEncodable:
            class ObjectExpr2403(IEncodable):
                def Encode(self, helpers_4: IEncoderHelpers_1[Any]) -> Any:
                    return helpers_4.encode_string(value_8)

            return ObjectExpr2403()

        def _arrow2406(value_10: str, person_1: Any=person_1) -> IEncodable:
            class ObjectExpr2405(IEncodable):
                def Encode(self, helpers_5: IEncoderHelpers_1[Any]) -> Any:
                    return helpers_5.encode_string(value_10)

            return ObjectExpr2405()

        def _arrow2408(value_12: str, person_1: Any=person_1) -> IEncodable:
            class ObjectExpr2407(IEncodable):
                def Encode(self, helpers_6: IEncoderHelpers_1[Any]) -> Any:
                    return helpers_6.encode_string(value_12)

            return ObjectExpr2407()

        def _arrow2410(value_14: str, person_1: Any=person_1) -> IEncodable:
            class ObjectExpr2409(IEncodable):
                def Encode(self, helpers_7: IEncoderHelpers_1[Any]) -> Any:
                    return helpers_7.encode_string(value_14)

            return ObjectExpr2409()

        def _arrow2412(value_16: str, person_1: Any=person_1) -> IEncodable:
            class ObjectExpr2411(IEncodable):
                def Encode(self, helpers_8: IEncoderHelpers_1[Any]) -> Any:
                    return helpers_8.encode_string(value_16)

            return ObjectExpr2411()

        def _arrow2413(oa: OntologyAnnotation, person_1: Any=person_1) -> IEncodable:
            return OntologyAnnotation_encoder(oa)

        def _arrow2414(comment: Comment, person_1: Any=person_1) -> IEncodable:
            return encoder_1(comment)

        values: FSharpList[tuple[str, IEncodable]] = choose(chooser, of_array([try_include("@id", _arrow2396, ROCrate_genID(person_2)), try_include("firstName", _arrow2398, person_2.FirstName), try_include("lastName", _arrow2400, person_2.LastName), try_include("midInitials", _arrow2402, person_2.MidInitials), try_include("email", _arrow2404, person_2.EMail), try_include("phone", _arrow2406, person_2.Phone), try_include("fax", _arrow2408, person_2.Fax), try_include("address", _arrow2410, person_2.Address), try_include("affiliation", _arrow2412, person_2.Affiliation), try_include_seq("roles", _arrow2413, person_2.Roles), try_include_seq("comments", _arrow2414, person_2.Comments)]))
        class ObjectExpr2415(IEncodable):
            def Encode(self, helpers_9: IEncoderHelpers_1[Any], person_1: Any=person_1) -> Any:
                def mapping_1(tupled_arg_1: tuple[str, IEncodable]) -> tuple[str, __A_]:
                    return (tupled_arg_1[0], tupled_arg_1[1].Encode(helpers_9))

                arg: IEnumerable_1[tuple[str, __A_]] = map_1(mapping_1, values)
                return helpers_9.encode_object(arg)

        return ObjectExpr2415()

    if id_map is not None:
        def _arrow2416(p_1: Person, id_map: Any=id_map, person: Any=person) -> str:
            return ROCrate_genID(p_1)

        return encode(_arrow2416, f, person, id_map)

    else: 
        return f(person)



def _arrow2427(get: IGetters) -> Person:
    def _arrow2417(__unit: None=None) -> str | None:
        object_arg: IOptionalGetter = get.Optional
        return object_arg.Field("lastName", string)

    def _arrow2418(__unit: None=None) -> str | None:
        object_arg_1: IOptionalGetter = get.Optional
        return object_arg_1.Field("firstName", string)

    def _arrow2419(__unit: None=None) -> str | None:
        object_arg_2: IOptionalGetter = get.Optional
        return object_arg_2.Field("midInitials", string)

    def _arrow2420(__unit: None=None) -> str | None:
        object_arg_3: IOptionalGetter = get.Optional
        return object_arg_3.Field("email", string)

    def _arrow2421(__unit: None=None) -> str | None:
        object_arg_4: IOptionalGetter = get.Optional
        return object_arg_4.Field("phone", string)

    def _arrow2422(__unit: None=None) -> str | None:
        object_arg_5: IOptionalGetter = get.Optional
        return object_arg_5.Field("fax", string)

    def _arrow2423(__unit: None=None) -> str | None:
        object_arg_6: IOptionalGetter = get.Optional
        return object_arg_6.Field("address", string)

    def _arrow2424(__unit: None=None) -> str | None:
        object_arg_7: IOptionalGetter = get.Optional
        return object_arg_7.Field("affiliation", string)

    def _arrow2425(__unit: None=None) -> Array[OntologyAnnotation] | None:
        arg_17: Decoder_1[Array[OntologyAnnotation]] = resize_array(OntologyAnnotation_decoder)
        object_arg_8: IOptionalGetter = get.Optional
        return object_arg_8.Field("roles", arg_17)

    def _arrow2426(__unit: None=None) -> Array[Comment] | None:
        arg_19: Decoder_1[Array[Comment]] = resize_array(decoder_1)
        object_arg_9: IOptionalGetter = get.Optional
        return object_arg_9.Field("comments", arg_19)

    return Person_setOrcidFromComments(Person(None, _arrow2417(), _arrow2418(), _arrow2419(), _arrow2420(), _arrow2421(), _arrow2422(), _arrow2423(), _arrow2424(), _arrow2425(), _arrow2426()))


ISAJson_decoder: Decoder_1[Person] = Decode_objectNoAdditionalProperties(ISAJson_allowedFields, _arrow2427)

__all__ = ["encoder", "decoder", "ROCrate_genID", "ROCrate_Affiliation_encoder", "ROCrate_Affiliation_decoder", "ROCrate_encoder", "ROCrate_decoder", "ROCrate_encodeAuthorListString", "ROCrate_decodeAuthorListString", "ISAJson_allowedFields", "ISAJson_encoder", "ISAJson_decoder"]




from __future__ import annotations

import builtins
import functools
import math
import platform
import random
import re
import weakref
from abc import ABC, abstractmethod
from array import array
from collections.abc import (
    Callable,
    Iterable,
    Iterator,
    MutableSequence,
    Sequence,
    Sized,
)
from contextlib import AbstractContextManager
from enum import IntEnum
from threading import RLock
from types import TracebackType
from typing import (
    Any,
    ClassVar,
    Generic,
    Protocol,
    TypeVar,
    cast,
)
from urllib.parse import quote, unquote


class SupportsLessThan(Protocol):
    @abstractmethod
    def __lt__(self, __other: Any) -> bool:
        raise NotImplementedError


_T = TypeVar("_T")
_T_in = TypeVar("_T_in", contravariant=True)
_T_out = TypeVar("_T_out", covariant=True)
_Key = TypeVar("_Key")
_Value = TypeVar("_Value")
_TSupportsLessThan = TypeVar("_TSupportsLessThan", bound=SupportsLessThan)

Array = MutableSequence


class ObjectDisposedException(Exception):
    def __init__(self):
        super().__init__("Cannot access a disposed object")


class IDisposable(ABC):
    """IDisposable interface.

    Note currently not a protocol since it also impmelents resource
    management and needs to be inherited from.
    """

    __slots__ = ()

    @abstractmethod
    def Dispose(self) -> None: ...

    def __enter__(self):
        """Enter context management."""
        return self

    def __exit__(
        self,
        exctype: type[BaseException] | None,
        excinst: BaseException | None,
        exctb: TracebackType | None,
    ) -> bool:
        """Exit context management."""

        self.Dispose()
        return False

    @staticmethod
    def create(action: Callable[[], None]):
        """Create disposable from action. Will call action when
        disposed."""
        return AnonymousDisposable(action)


Disposable = TypeVar("Disposable", bound=IDisposable)


class AnonymousDisposable(IDisposable):
    __slots__ = "_is_disposed", "_action", "_lock"

    def __init__(self, action: Callable[[], None]):
        self._is_disposed = False
        self._action = action
        self._lock = RLock()

    def Dispose(self) -> None:
        """Performs the task of cleaning up resources."""

        dispose = False
        with self._lock:
            if not self._is_disposed:
                dispose = True
                self._is_disposed = True

        if dispose:
            self._action()

    def __enter__(self):
        if self._is_disposed:
            raise ObjectDisposedException()
        return self


class IEquatable(Protocol):
    @abstractmethod
    def __eq__(self, other: Any) -> bool:
        raise NotImplementedError

    @abstractmethod
    def __hash__(self) -> int:
        raise NotImplementedError


class IComparable(IEquatable, Protocol):
    @abstractmethod
    def __cmp__(self, __other: Any) -> int:
        raise NotImplementedError

    @abstractmethod
    def __lt__(self, other: Any) -> bool:
        raise NotImplementedError


class IComparable_1(Generic[_T_in], IEquatable, Protocol):
    @abstractmethod
    def __cmp__(self, __other: _T_in) -> int:
        raise NotImplementedError

    @abstractmethod
    def __lt__(self, other: Any) -> bool:
        raise NotImplementedError


class IComparer(Protocol):
    """Defines a method that a type implements to compare two objects.

    https://docs.microsoft.com/en-us/dotnet/api/system.collections.generic.icomparer-1
    """

    @property
    @abstractmethod
    def Compare(self) -> Callable[[_T_in, _T_in], int]: ...


class IComparer_1(Generic[_T_in], Protocol):
    """Defines a method that a type implements to compare two objects.

    https://docs.microsoft.com/en-us/dotnet/api/system.collections.generic.icomparer-1
    """

    @property
    @abstractmethod
    def Compare(self) -> Callable[[_T_in, _T_in], int]: ...


class IEqualityComparer(Protocol):
    def Equals(self, __x: _T_in, __y: _T_in) -> bool:
        return self.System_Collections_IEqualityComparer_Equals541DA560(__x, __y)

    def GetHashCode(self) -> int:
        return self.System_Collections_IEqualityComparer_GetHashCode4E60E31B()

    @abstractmethod
    def System_Collections_IEqualityComparer_Equals541DA560(self, x: Any = None, y: Any = None) -> bool:
        raise NotImplementedError

    @abstractmethod
    def System_Collections_IEqualityComparer_GetHashCode4E60E31B(self, x_1: Any = None) -> int:
        raise NotImplementedError


class IEqualityComparer_1(Generic[_T_in], Protocol):
    def Equals(self, __x: _T_in, __y: _T_in) -> bool:
        raise NotImplementedError

    def GetHashCode(self) -> int:
        raise NotImplementedError


class IStructuralEquatable(Protocol):
    @abstractmethod
    def Equals(self, other: Any, comparer: IEqualityComparer) -> bool:
        return NotImplemented

    @abstractmethod
    def __hash__(self) -> int:
        raise NotImplementedError


class IStructuralComparable(Protocol):
    @abstractmethod
    def __cmp__(self, other: Any, comparer: IComparer) -> int:
        raise NotImplementedError


class DateKind(IntEnum):
    Unspecified = 0
    UTC = 1
    Local = 2


def equals(a: Any, b: Any) -> bool:
    match (a, b):
        case (a, b) if a is b:
            return True
        # Don't test (None, None) here, because a is b already covers that
        # case (None, None):
        #     return True
        case (None, _):
            return False
        case (_, None):
            return False
        case (a, b) if is_array_like(a):
            return equal_arrays(a, b)
        case _:
            return a == b


def is_comparable(x: Any) -> bool:
    return hasattr(x, "__cmp__") and callable(x.__cmp__)


def is_equatable(x: Any) -> bool:
    return hasattr(x, "__eq__") and callable(x.__eq__)


def is_iterable(x: Any) -> bool:
    return isinstance(x, Iterable)


def compare_dicts(x: dict[str, Any], y: dict[str, Any]) -> int:
    """Compare Python dicts with string keys.

    Python cannot do this natively.
    """
    x_keys = x.keys()
    y_keys = y.keys()

    if len(x_keys) != len(y_keys):
        return -1 if len(x_keys) < len(y_keys) else 1

    x_keys = sorted(x_keys)
    y_keys = sorted(y_keys)

    j = 0
    for i, key in enumerate(x_keys):
        if key != y_keys[i]:
            return -1 if key < y_keys[i] else 1
        else:
            j = compare(x[key], y[key])
            if j != 0:
                return j

    return 0


def compare_arrays(xs: list[Any] | None, ys: list[Any] | None):
    if xs is None:
        return 0 if ys is None else 1

    if ys is None:
        return -1

    if len(xs) != len(ys):
        return -1 if len(xs) < len(ys) else 1

    j = 0
    for i, x in enumerate(xs):
        j = compare(x, ys[i])
        if j != 0:
            return j

    return 0


def compare(a: Any, b: Any) -> int:
    if a is b:
        return 0

    if a is None:
        return -1 if b else 0

    if b is None:
        return 1 if a else 0

    if is_comparable(a):
        return a.__cmp__(b)

    if isinstance(a, dict):
        return compare_dicts(cast(dict[str, Any], a), b)

    if isinstance(a, list):
        return compare_arrays(cast(list[Any], a), b)

    if is_equatable(a) and a == b:
        return 0

    if hasattr(a, "__lt__") and callable(a.__lt__) and a < b:
        return -1

    return 1


def equal_arrays_with(xs: Sequence[_T] | None, ys: Sequence[_T] | None, eq: Callable[[_T, _T], bool]) -> bool:
    if xs is None:
        return ys is None

    if ys is None:
        return False

    if len(xs) != len(ys):
        return False

    for i, x in enumerate(xs):
        if not eq(x, ys[i]):
            return False

    return True


def equal_arrays(x: Sequence[_T], y: Sequence[_T]) -> bool:
    return equal_arrays_with(x, y, equals)


def compare_primitives(x: _TSupportsLessThan, y: _TSupportsLessThan) -> int:
    return 0 if x == y else (-1 if x < y else 1)


def min(comparer: Callable[[_T, _T], int], x: _T, y: _T) -> _T:
    return x if comparer(x, y) < 0 else y


def max(comparer: Callable[[_T, _T], int], x: _T, y: _T) -> _T:
    return x if comparer(x, y) > 0 else y


def clamp(comparer: Callable[[_T, _T], int], value: _T, min: _T, max: _T):
    return min if (comparer(value, min) < 0) else max if comparer(value, max) > 0 else value


def assert_equal(actual: Any, expected: Any, msg: str | None = None) -> None:
    if not equals(actual, expected):
        raise Exception(msg or f"Expected: {expected} - Actual: {actual}")


def assert_not_equal(actual: _T, expected: _T, msg: str | None = None) -> None:
    if equals(actual, expected):
        raise Exception(msg or f"Expected: {expected} - Actual: {actual}")


MAX_LOCKS = 1024


def lock(lock_obj: Any, fn: Callable[[], _T]) -> _T:
    @functools.lru_cache(maxsize=MAX_LOCKS)
    def get_lock(n: int) -> RLock:
        return RLock()

    lock = get_lock(id(lock_obj))
    with lock:
        return fn()


class Lazy(Generic[_T]):
    def __init__(self, factory: Callable[[], _T]):
        self.factory = factory
        self.is_value_created: bool = False
        self.created_value: _T | None = None

    @property
    def Value(self) -> _T:
        if not self.is_value_created:
            self.created_value = self.factory()
            self.is_value_created = True

        assert self.created_value is not None
        return self.created_value

    @property
    def IsValueCreated(self):
        return self.is_value_created


def lazy_from_value(v: _T) -> Lazy[_T]:
    return Lazy(lambda: v)


def pad_with_zeros(i: int, length: int) -> str:
    string = str(i)
    while len(string) < length:
        string = "0" + string
    return string


def pad_left_and_right_with_zeros(i: int, length_left: int, length_right: int) -> str:
    string = str(i)
    while len(string) < length_left:
        string = "0" + string
    while len(string) < length_right:
        string = string + "0"
    return string


class Atom(Generic[_T], Protocol):
    def __call__(self, *value: _T) -> _T | None: ...


def create_atom(value: _T) -> Atom[_T]:
    atom = value

    def wrapper(*value: _T) -> _T | None:
        nonlocal atom

        if len(value) == 0:
            return atom

        atom = value[0]
        return None

    return wrapper


def create_obj(fields: list[tuple[Any, Any]]):
    return dict(fields)


def tohex(val: int, nbits: int | None = None) -> str:
    if nbits:
        val = (val + (1 << nbits)) % (1 << nbits)
    return f"{val:x}"


def int_to_string(i: int, radix: int = 10, bitsize: int | None = None) -> str:
    if radix == 10:
        return f"{i:d}"
    if radix == 16:
        return tohex(i, bitsize)
    if radix == 2:
        return f"{i:b}"
    if radix == 8:
        return f"{i:o}"
    return str(i)


def int8_to_string(i: int, radix: int = 10, bitsize: int | None = None) -> str:
    return int_to_string(i, radix, 8)


def int16_to_string(i: int, radix: int = 10, bitsize: int | None = None) -> str:
    return int_to_string(i, radix, 16)


def int32_to_string(i: int, radix: int = 10, bitsize: int | None = None) -> str:
    return int_to_string(i, radix, 32)


def int64_to_string(i: int, radix: int = 10, bitsize: int | None = None) -> str:
    return int_to_string(i, radix, 64)


def count(col: Iterable[Any]) -> int:
    if isinstance(col, Sized):
        return len(col)

    count = 0
    for _ in col:
        count += 1

    return count


def clear(col: dict[Any, Any] | list[Any] | None) -> None:
    if isinstance(col, list | dict):
        col.clear()


class IEnumerator(Iterator[_T], IDisposable):
    __slots__ = ()

    def Current(self) -> _T:
        return self.System_Collections_Generic_IEnumerator_1_get_Current()

    def MoveNext(self) -> bool:
        return self.System_Collections_IEnumerator_MoveNext()

    def Reset(self) -> None:
        return self.System_Collections_IEnumerator_Reset()

    @abstractmethod
    def System_Collections_Generic_IEnumerator_1_get_Current(self) -> _T: ...

    def System_Collections_IEnumerator_get_Current(self) -> Any:
        return self.System_Collections_Generic_IEnumerator_1_get_Current()

    @abstractmethod
    def System_Collections_IEnumerator_MoveNext(self) -> bool: ...

    @abstractmethod
    def System_Collections_IEnumerator_Reset(self) -> None: ...

    def __iter__(self) -> Iterator[_T]:
        return self

    def __next__(self) -> _T:
        if not self.MoveNext():
            raise StopIteration
        return self.Current()


class IEnumerable(Iterable[Any], Protocol):
    __slots__ = ()

    @abstractmethod
    def GetEnumerator(self) -> IEnumerator[Any]: ...

    def __iter__(self) -> Iterator[Any]:
        return self.GetEnumerator()


class IEnumerable_1(Iterable[_T], Protocol):
    __slots__ = ()

    @abstractmethod
    def GetEnumerator(self) -> IEnumerator[_T]: ...

    def __iter__(self) -> Iterator[_T]:
        return self.GetEnumerator()


class ICollection(IEnumerable_1[_T], Protocol): ...


class IDictionary(ICollection[tuple[_Key, _Value]], Protocol):
    @abstractmethod
    def keys(self) -> IEnumerable_1[_Key]: ...

    def values(self) -> IEnumerable_1[_Value]: ...


class Enumerator(IEnumerator[_T]):
    __slots__ = "iter", "current"

    def __init__(self, iter: Iterator[_T]) -> None:
        self.iter = iter
        self.current = None

    def System_Collections_Generic_IEnumerator_1_get_Current(self) -> _T:
        if self.current is not None:
            return self.current
        return None  # type: ignore

    def System_Collections_IEnumerator_MoveNext(self) -> bool:
        try:
            cur = next(self.iter)
            self.current = cur
            return True
        except StopIteration:
            return False

    def System_Collections_IEnumerator_Reset(self) -> None:
        raise Exception("Python iterators cannot be reset")

    def Dispose(self) -> None:
        return

    def __next__(self) -> _T:
        return next(self.iter)

    def __iter__(self) -> Iterator[_T]:
        return self


class Enumerable(IEnumerable_1[_T]):
    __slots__ = "xs"

    def __init__(self, xs: Iterable[_T]) -> None:
        self.xs = xs

    def GetEnumerator(self) -> IEnumerator[_T]:
        return Enumerator(iter(self.xs))

    def __iter__(self) -> Iterator[_T]:
        return iter(self.xs)


def to_enumerable(e: Iterable[_T]) -> IEnumerable_1[_T]:
    return Enumerable(e)


def get_enumerator(o: Iterable[Any]) -> Enumerator[Any]:
    attr = getattr(o, "GetEnumerator", None)
    if attr:
        return attr()
    elif isinstance(o, dict):
        # Dictionaries should produce tuples
        return Enumerator(iter(cast(Any, o.items())))
    else:
        return Enumerator(iter(o))


_T1 = TypeVar("_T1")
_T2 = TypeVar("_T2")
_T3 = TypeVar("_T3")
_T4 = TypeVar("_T4")
_T5 = TypeVar("_T5")
_T6 = TypeVar("_T6")
_T7 = TypeVar("_T7")
_T8 = TypeVar("_T8")
_T9 = TypeVar("_T9")
_T10 = TypeVar("_T10")
_T11 = TypeVar("_T11")
_T12 = TypeVar("_T12")
_T13 = TypeVar("_T13")
_T14 = TypeVar("_T14")
_T15 = TypeVar("_T15")
_T16 = TypeVar("_T16")
_T17 = TypeVar("_T17")
_T18 = TypeVar("_T18")
_T19 = TypeVar("_T19")
_T20 = TypeVar("_T20")
_TResult = TypeVar("_TResult")

_curried = weakref.WeakKeyDictionary[Any, Any]()


def uncurry2(f: Callable[[_T1], Callable[[_T2], _TResult]]) -> Callable[[_T1, _T2], _TResult]:
    def f2(a1: _T1, a2: _T2) -> _TResult:
        return f(a1)(a2)

    _curried[f2] = f
    return f2


def curry2(f: Callable[[_T1, _T2], _TResult]) -> Callable[[_T1], Callable[[_T2], _TResult]]:
    f2 = _curried.get(f)
    if f2 is None:
        return lambda a1: lambda a2: f(a1, a2)
    else:
        return f2


def uncurry3(f: Callable[[_T1], Callable[[_T2], Callable[[_T3], _TResult]]]) -> Callable[[_T1, _T2, _T3], _TResult]:
    def f2(a1: _T1, a2: _T2, a3: _T3) -> _TResult:
        return f(a1)(a2)(a3)

    _curried[f2] = f
    return f2


def curry3(f: Callable[[_T1, _T2, _T3], _TResult]) -> Callable[[_T1], Callable[[_T2], Callable[[_T3], _TResult]]]:
    f2 = _curried.get(f)
    if f2 is None:
        return lambda a1: lambda a2: lambda a3: f(a1, a2, a3)
    else:
        return f2


def uncurry4(
    f: Callable[[_T1], Callable[[_T2], Callable[[_T3], Callable[[_T4], _TResult]]]],
) -> Callable[[_T1, _T2, _T3, _T4], _TResult]:
    def f2(a1: _T1, a2: _T2, a3: _T3, a4: _T4) -> _TResult:
        return f(a1)(a2)(a3)(a4)

    _curried[f2] = f
    return f2


def curry4(
    f: Callable[[_T1, _T2, _T3, _T4], _TResult],
) -> Callable[[_T1], Callable[[_T2], Callable[[_T3], Callable[[_T4], _TResult]]]]:
    f2 = _curried.get(f)
    if f2 is None:
        return lambda a1: lambda a2: lambda a3: lambda a4: f(a1, a2, a3, a4)
    else:
        return f2


def uncurry5(
    f: Callable[
        [_T1],
        Callable[[_T2], Callable[[_T3], Callable[[_T4], Callable[[_T5], _TResult]]]],
    ],
) -> Callable[[_T1, _T2, _T3, _T4, _T5], _TResult]:
    def f2(a1: _T1, a2: _T2, a3: _T3, a4: _T4, a5: _T5) -> _TResult:
        return f(a1)(a2)(a3)(a4)(a5)

    _curried[f2] = f
    return f2


def curry5(
    f: Callable[[_T1, _T2, _T3, _T4, _T5], _TResult],
) -> Callable[[_T1], Callable[[_T2], Callable[[_T3], Callable[[_T4], Callable[[_T5], _TResult]]]]]:
    f2 = _curried.get(f)
    if f2 is None:
        return lambda a1: lambda a2: lambda a3: lambda a4: lambda a5: f(a1, a2, a3, a4, a5)
    else:
        return f2


def uncurry6(
    f: Callable[
        [_T1],
        Callable[
            [_T2],
            Callable[[_T3], Callable[[_T4], Callable[[_T5], Callable[[_T6], _TResult]]]],
        ],
    ],
) -> Callable[[_T1, _T2, _T3, _T4, _T5, _T6], _TResult]:
    def f2(a1: _T1, a2: _T2, a3: _T3, a4: _T4, a5: _T5, a6: _T6) -> _TResult:
        return f(a1)(a2)(a3)(a4)(a5)(a6)

    _curried[f2] = f
    return f2


def curry6(
    f: Callable[[_T1, _T2, _T3, _T4, _T5, _T6], _TResult],
) -> Callable[
    [_T1],
    Callable[
        [_T2],
        Callable[[_T3], Callable[[_T4], Callable[[_T5], Callable[[_T6], _TResult]]]],
    ],
]:
    f2 = _curried.get(f)
    if f2 is None:
        return lambda a1: lambda a2: lambda a3: lambda a4: lambda a5: lambda a6: f(a1, a2, a3, a4, a5, a6)
    else:
        return f2


def uncurry7(
    f: Callable[
        [_T1],
        Callable[
            [_T2],
            Callable[
                [_T3],
                Callable[[_T4], Callable[[_T5], Callable[[_T6], Callable[[_T7], _TResult]]]],
            ],
        ],
    ],
) -> Callable[[_T1, _T2, _T3, _T4, _T5, _T6, _T7], _TResult]:
    def f2(a1: _T1, a2: _T2, a3: _T3, a4: _T4, a5: _T5, a6: _T6, a7: _T7) -> _TResult:
        return f(a1)(a2)(a3)(a4)(a5)(a6)(a7)

    _curried[f2] = f
    return f2


def curry7(
    f: Callable[[_T1, _T2, _T3, _T4, _T5, _T6, _T7], _TResult],
) -> Callable[
    [_T1],
    Callable[
        [_T2],
        Callable[
            [_T3],
            Callable[[_T4], Callable[[_T5], Callable[[_T6], Callable[[_T7], _TResult]]]],
        ],
    ],
]:
    f2 = _curried.get(f)
    if f2 is None:
        return lambda a1: lambda a2: lambda a3: lambda a4: lambda a5: lambda a6: lambda a7: f(
            a1, a2, a3, a4, a5, a6, a7
        )
    else:
        return f2


def uncurry8(
    f: Callable[
        [_T1],
        Callable[
            [_T2],
            Callable[
                [_T3],
                Callable[
                    [_T4],
                    Callable[
                        [_T5],
                        Callable[[_T6], Callable[[_T7], Callable[[_T8], _TResult]]],
                    ],
                ],
            ],
        ],
    ],
) -> Callable[[_T1, _T2, _T3, _T4, _T5, _T6, _T7, _T8], _TResult]:
    def f2(a1: _T1, a2: _T2, a3: _T3, a4: _T4, a5: _T5, a6: _T6, a7: _T7, a8: _T8) -> _TResult:
        return f(a1)(a2)(a3)(a4)(a5)(a6)(a7)(a8)

    _curried[f2] = f
    return f2


def curry8(
    f: Callable[[_T1, _T2, _T3, _T4, _T5, _T6, _T7, _T8], _TResult],
) -> Callable[
    [_T1],
    Callable[
        [_T2],
        Callable[
            [_T3],
            Callable[
                [_T4],
                Callable[[_T5], Callable[[_T6], Callable[[_T7], Callable[[_T8], _TResult]]]],
            ],
        ],
    ],
]:
    f2 = _curried.get(f)
    if f2 is None:
        return lambda a1: lambda a2: lambda a3: lambda a4: lambda a5: lambda a6: lambda a7: lambda a8: f(
            a1, a2, a3, a4, a5, a6, a7, a8
        )
    else:
        return f2


def uncurry9(
    f: Callable[
        [_T1],
        Callable[
            [_T2],
            Callable[
                [_T3],
                Callable[
                    [_T4],
                    Callable[
                        [_T5],
                        Callable[
                            [_T6],
                            Callable[[_T7], Callable[[_T8], Callable[[_T9], _TResult]]],
                        ],
                    ],
                ],
            ],
        ],
    ],
) -> Callable[[_T1, _T2, _T3, _T4, _T5, _T6, _T7, _T8, _T9], _TResult]:
    def f2(a1: _T1, a2: _T2, a3: _T3, a4: _T4, a5: _T5, a6: _T6, a7: _T7, a8: _T8, a9: _T9) -> _TResult:
        return f(a1)(a2)(a3)(a4)(a5)(a6)(a7)(a8)(a9)

    _curried[f2] = f
    return f2


def curry9(
    f: Callable[[_T1, _T2, _T3, _T4, _T5, _T6, _T7, _T8, _T9], _TResult],
) -> Callable[
    [_T1],
    Callable[
        [_T2],
        Callable[
            [_T3],
            Callable[
                [_T4],
                Callable[
                    [_T5],
                    Callable[
                        [_T6],
                        Callable[[_T7], Callable[[_T8], Callable[[_T9], _TResult]]],
                    ],
                ],
            ],
        ],
    ],
]:
    f2 = _curried.get(f)
    if f2 is None:
        return lambda a1: lambda a2: lambda a3: lambda a4: lambda a5: lambda a6: lambda a7: lambda a8: lambda a9: f(
            a1, a2, a3, a4, a5, a6, a7, a8, a9
        )
    else:
        return f2


def uncurry10(
    f: Callable[
        [_T1],
        Callable[
            [_T2],
            Callable[
                [_T3],
                Callable[
                    [_T4],
                    Callable[
                        [_T5],
                        Callable[
                            [_T6],
                            Callable[
                                [_T7],
                                Callable[[_T8], Callable[[_T9], Callable[[_T10], _TResult]]],
                            ],
                        ],
                    ],
                ],
            ],
        ],
    ],
) -> Callable[[_T1, _T2, _T3, _T4, _T5, _T6, _T7, _T8, _T9, _T10], _TResult]:
    def f2(
        a1: _T1,
        a2: _T2,
        a3: _T3,
        a4: _T4,
        a5: _T5,
        a6: _T6,
        a7: _T7,
        a8: _T8,
        a9: _T9,
        a10: _T10,
    ) -> _TResult:
        return f(a1)(a2)(a3)(a4)(a5)(a6)(a7)(a8)(a9)(a10)

    _curried[f2] = f
    return f2


def curry10(
    f: Callable[[_T1, _T2, _T3, _T4, _T5, _T6, _T7, _T8, _T9, _T10], _TResult],
) -> Callable[
    [_T1],
    Callable[
        [_T2],
        Callable[
            [_T3],
            Callable[
                [_T4],
                Callable[
                    [_T5],
                    Callable[
                        [_T6],
                        Callable[
                            [_T7],
                            Callable[[_T8], Callable[[_T9], Callable[[_T10], _TResult]]],
                        ],
                    ],
                ],
            ],
        ],
    ],
]:
    f2 = _curried.get(f)
    if f2 is None:
        return (
            lambda a1: lambda a2: lambda a3: lambda a4: lambda a5: lambda a6: lambda a7: lambda a8: lambda a9: lambda a10: f(
                a1, a2, a3, a4, a5, a6, a7, a8, a9, a10
            )
        )
    else:
        return f2


def uncurry11(
    f: Callable[
        [_T1],
        Callable[
            [_T2],
            Callable[
                [_T3],
                Callable[
                    [_T4],
                    Callable[
                        [_T5],
                        Callable[
                            [_T6],
                            Callable[
                                [_T7],
                                Callable[
                                    [_T8],
                                    Callable[
                                        [_T9],
                                        Callable[[_T10], Callable[[_T11], _TResult]],
                                    ],
                                ],
                            ],
                        ],
                    ],
                ],
            ],
        ],
    ],
) -> Callable[[_T1, _T2, _T3, _T4, _T5, _T6, _T7, _T8, _T9, _T10, _T11], _TResult]:
    def f2(
        a1: Any,
        a2: Any,
        a3: Any,
        a4: Any,
        a5: Any,
        a6: Any,
        a7: Any,
        a8: Any,
        a9: Any,
        a10: Any,
        a11: Any,
    ) -> Any:
        return f(a1)(a2)(a3)(a4)(a5)(a6)(a7)(a8)(a9)(a10)(a11)

    _curried[f2] = f
    return f2


def curry11(
    f: Callable[[_T1, _T2, _T3, _T4, _T5, _T6, _T7, _T8, _T9, _T10, _T11], _TResult],
) -> Callable[
    [_T1],
    Callable[
        [_T2],
        Callable[
            [_T3],
            Callable[
                [_T4],
                Callable[
                    [_T5],
                    Callable[
                        [_T6],
                        Callable[
                            [_T7],
                            Callable[
                                [_T8],
                                Callable[[_T9], Callable[[_T10], Callable[[_T11], _TResult]]],
                            ],
                        ],
                    ],
                ],
            ],
        ],
    ],
]:
    f2 = _curried.get(f)
    if f2 is None:
        return (
            lambda a1: lambda a2: lambda a3: lambda a4: lambda a5: lambda a6: lambda a7: lambda a8: lambda a9: lambda a10: lambda a11: f(
                a1, a2, a3, a4, a5, a6, a7, a8, a9, a10, a11
            )
        )
    else:
        return f2


def uncurry12(
    f: Callable[
        [_T1],
        Callable[
            [_T2],
            Callable[
                [_T3],
                Callable[
                    [_T4],
                    Callable[
                        [_T5],
                        Callable[
                            [_T6],
                            Callable[
                                [_T7],
                                Callable[
                                    [_T8],
                                    Callable[
                                        [_T9],
                                        Callable[
                                            [_T10],
                                            Callable[[_T11], Callable[[_T12], _TResult]],
                                        ],
                                    ],
                                ],
                            ],
                        ],
                    ],
                ],
            ],
        ],
    ],
) -> Callable[[_T1, _T2, _T3, _T4, _T5, _T6, _T7, _T8, _T9, _T10, _T11, _T12], _TResult]:
    def f2(
        a1: Any,
        a2: Any,
        a3: Any,
        a4: Any,
        a5: Any,
        a6: Any,
        a7: Any,
        a8: Any,
        a9: Any,
        a10: Any,
        a11: Any,
        a12: Any,
    ) -> Any:
        return f(a1)(a2)(a3)(a4)(a5)(a6)(a7)(a8)(a9)(a10)(a11)(a12)

    _curried[f2] = f
    return f2


def curry12(
    f: Callable[[_T1, _T2, _T3, _T4, _T5, _T6, _T7, _T8, _T9, _T10, _T11, _T12], _TResult],
) -> Callable[
    [_T1],
    Callable[
        [_T2],
        Callable[
            [_T3],
            Callable[
                [_T4],
                Callable[
                    [_T5],
                    Callable[
                        [_T6],
                        Callable[
                            [_T7],
                            Callable[
                                [_T8],
                                Callable[
                                    [_T9],
                                    Callable[
                                        [_T10],
                                        Callable[[_T11], Callable[[_T12], _TResult]],
                                    ],
                                ],
                            ],
                        ],
                    ],
                ],
            ],
        ],
    ],
]:
    f2 = _curried.get(f)
    if f2 is None:
        return (
            lambda a1: lambda a2: lambda a3: lambda a4: lambda a5: lambda a6: lambda a7: lambda a8: lambda a9: lambda a10: lambda a11: lambda a12: f(
                a1, a2, a3, a4, a5, a6, a7, a8, a9, a10, a11, a12
            )
        )
    else:
        return f2


def uncurry13(
    f: Callable[
        [_T1],
        Callable[
            [_T2],
            Callable[
                [_T3],
                Callable[
                    [_T4],
                    Callable[
                        [_T5],
                        Callable[
                            [_T6],
                            Callable[
                                [_T7],
                                Callable[
                                    [_T8],
                                    Callable[
                                        [_T9],
                                        Callable[
                                            [_T10],
                                            Callable[
                                                [_T11],
                                                Callable[[_T12], Callable[[_T13], _TResult]],
                                            ],
                                        ],
                                    ],
                                ],
                            ],
                        ],
                    ],
                ],
            ],
        ],
    ],
) -> Callable[[_T1, _T2, _T3, _T4, _T5, _T6, _T7, _T8, _T9, _T10, _T11, _T12, _T13], _TResult]:
    def f2(
        a1: Any,
        a2: Any,
        a3: Any,
        a4: Any,
        a5: Any,
        a6: Any,
        a7: Any,
        a8: Any,
        a9: Any,
        a10: Any,
        a11: Any,
        a12: Any,
        a13: Any,
    ) -> Any:
        return f(a1)(a2)(a3)(a4)(a5)(a6)(a7)(a8)(a9)(a10)(a11)(a12)(a13)

    _curried[f2] = f
    return f2


def curry13(
    f: Callable[[_T1, _T2, _T3, _T4, _T5, _T6, _T7, _T8, _T9, _T10, _T11, _T12, _T13], _TResult],
) -> Callable[
    [_T1],
    Callable[
        [_T2],
        Callable[
            [_T3],
            Callable[
                [_T4],
                Callable[
                    [_T5],
                    Callable[
                        [_T6],
                        Callable[
                            [_T7],
                            Callable[
                                [_T8],
                                Callable[
                                    [_T9],
                                    Callable[
                                        [_T10],
                                        Callable[
                                            [_T11],
                                            Callable[[_T12], Callable[[_T13], _TResult]],
                                        ],
                                    ],
                                ],
                            ],
                        ],
                    ],
                ],
            ],
        ],
    ],
]:
    f2 = _curried.get(f)
    if f2 is None:
        return (
            lambda a1: lambda a2: lambda a3: lambda a4: lambda a5: lambda a6: lambda a7: lambda a8: lambda a9: lambda a10: lambda a11: lambda a12: lambda a13: f(
                a1, a2, a3, a4, a5, a6, a7, a8, a9, a10, a11, a12, a13
            )
        )
    else:
        return f2


def uncurry14(
    f: Callable[
        [_T1],
        Callable[
            [_T2],
            Callable[
                [_T3],
                Callable[
                    [_T4],
                    Callable[
                        [_T5],
                        Callable[
                            [_T6],
                            Callable[
                                [_T7],
                                Callable[
                                    [_T8],
                                    Callable[
                                        [_T9],
                                        Callable[
                                            [_T10],
                                            Callable[
                                                [_T11],
                                                Callable[
                                                    [_T12],
                                                    Callable[
                                                        [_T13],
                                                        Callable[[_T14], _TResult],
                                                    ],
                                                ],
                                            ],
                                        ],
                                    ],
                                ],
                            ],
                        ],
                    ],
                ],
            ],
        ],
    ],
) -> Callable[
    [
        _T1,
        _T2,
        _T3,
        _T4,
        _T5,
        _T6,
        _T7,
        _T8,
        _T9,
        _T10,
        _T11,
        _T12,
        _T13,
        _T14,
    ],
    _TResult,
]:
    def f2(
        a1: Any,
        a2: Any,
        a3: Any,
        a4: Any,
        a5: Any,
        a6: Any,
        a7: Any,
        a8: Any,
        a9: Any,
        a10: Any,
        a11: Any,
        a12: Any,
        a13: Any,
        a14: Any,
    ) -> Any:
        return f(a1)(a2)(a3)(a4)(a5)(a6)(a7)(a8)(a9)(a10)(a11)(a12)(a13)(a14)

    _curried[f2] = f
    return f2


def curry14(
    f: Callable[
        [_T1, _T2, _T3, _T4, _T5, _T6, _T7, _T8, _T9, _T10, _T11, _T12, _T13, _T14],
        _TResult,
    ],
) -> Callable[
    [_T1],
    Callable[
        [_T2],
        Callable[
            [_T3],
            Callable[
                [_T4],
                Callable[
                    [_T5],
                    Callable[
                        [_T6],
                        Callable[
                            [_T7],
                            Callable[
                                [_T8],
                                Callable[
                                    [_T9],
                                    Callable[
                                        [_T10],
                                        Callable[
                                            [_T11],
                                            Callable[
                                                [_T12],
                                                Callable[[_T13], Callable[[_T14], _TResult]],
                                            ],
                                        ],
                                    ],
                                ],
                            ],
                        ],
                    ],
                ],
            ],
        ],
    ],
]:
    f2 = _curried.get(f)
    if f2 is None:
        return (
            lambda a1: lambda a2: lambda a3: lambda a4: lambda a5: lambda a6: lambda a7: lambda a8: lambda a9: lambda a10: lambda a11: lambda a12: lambda a13: lambda a14: f(
                a1, a2, a3, a4, a5, a6, a7, a8, a9, a10, a11, a12, a13, a14
            )
        )
    else:
        return f2


def uncurry15(
    f: Callable[
        [_T1],
        Callable[
            [_T2],
            Callable[
                [_T3],
                Callable[
                    [_T4],
                    Callable[
                        [_T5],
                        Callable[
                            [_T6],
                            Callable[
                                [_T7],
                                Callable[
                                    [_T8],
                                    Callable[
                                        [_T9],
                                        Callable[
                                            [_T10],
                                            Callable[
                                                [_T11],
                                                Callable[
                                                    [_T12],
                                                    Callable[
                                                        [_T13],
                                                        Callable[
                                                            [_T14],
                                                            Callable[[_T15], _TResult],
                                                        ],
                                                    ],
                                                ],
                                            ],
                                        ],
                                    ],
                                ],
                            ],
                        ],
                    ],
                ],
            ],
        ],
    ],
) -> Callable[
    [_T1, _T2, _T3, _T4, _T5, _T6, _T7, _T8, _T9, _T10, _T11, _T12, _T13, _T14, _T15],
    _TResult,
]:
    def f2(
        a1: Any,
        a2: Any,
        a3: Any,
        a4: Any,
        a5: Any,
        a6: Any,
        a7: Any,
        a8: Any,
        a9: Any,
        a10: Any,
        a11: Any,
        a12: Any,
        a13: Any,
        a14: Any,
        a15: Any,
    ) -> Any:
        return f(a1)(a2)(a3)(a4)(a5)(a6)(a7)(a8)(a9)(a10)(a11)(a12)(a13)(a14)(a15)

    _curried[f2] = f
    return f2


def curry15(
    f: Callable[
        [
            _T1,
            _T2,
            _T3,
            _T4,
            _T5,
            _T6,
            _T7,
            _T8,
            _T9,
            _T10,
            _T11,
            _T12,
            _T13,
            _T14,
            _T15,
        ],
        _TResult,
    ],
) -> Callable[
    [_T1],
    Callable[
        [_T2],
        Callable[
            [_T3],
            Callable[
                [_T4],
                Callable[
                    [_T5],
                    Callable[
                        [_T6],
                        Callable[
                            [_T7],
                            Callable[
                                [_T8],
                                Callable[
                                    [_T9],
                                    Callable[
                                        [_T10],
                                        Callable[
                                            [_T11],
                                            Callable[
                                                [_T12],
                                                Callable[
                                                    [_T13],
                                                    Callable[
                                                        [_T14],
                                                        Callable[[_T15], _TResult],
                                                    ],
                                                ],
                                            ],
                                        ],
                                    ],
                                ],
                            ],
                        ],
                    ],
                ],
            ],
        ],
    ],
]:
    f2 = _curried.get(f)
    if f2 is None:
        return (
            lambda a1: lambda a2: lambda a3: lambda a4: lambda a5: lambda a6: lambda a7: lambda a8: lambda a9: lambda a10: lambda a11: lambda a12: lambda a13: lambda a14: lambda a15: f(
                a1, a2, a3, a4, a5, a6, a7, a8, a9, a10, a11, a12, a13, a14, a15
            )
        )
    else:
        return f2


def uncurry16(
    f: Callable[
        [_T1],
        Callable[
            [_T2],
            Callable[
                [_T3],
                Callable[
                    [_T4],
                    Callable[
                        [_T5],
                        Callable[
                            [_T6],
                            Callable[
                                [_T7],
                                Callable[
                                    [_T8],
                                    Callable[
                                        [_T9],
                                        Callable[
                                            [_T10],
                                            Callable[
                                                [_T11],
                                                Callable[
                                                    [_T12],
                                                    Callable[
                                                        [_T13],
                                                        Callable[
                                                            [_T14],
                                                            Callable[
                                                                [_T15],
                                                                Callable[[_T16], _TResult],
                                                            ],
                                                        ],
                                                    ],
                                                ],
                                            ],
                                        ],
                                    ],
                                ],
                            ],
                        ],
                    ],
                ],
            ],
        ],
    ],
) -> Callable[
    [
        _T1,
        _T2,
        _T3,
        _T4,
        _T5,
        _T6,
        _T7,
        _T8,
        _T9,
        _T10,
        _T11,
        _T12,
        _T13,
        _T14,
        _T15,
        _T16,
    ],
    _TResult,
]:
    def f2(
        a1: Any,
        a2: Any,
        a3: Any,
        a4: Any,
        a5: Any,
        a6: Any,
        a7: Any,
        a8: Any,
        a9: Any,
        a10: Any,
        a11: Any,
        a12: Any,
        a13: Any,
        a14: Any,
        a15: Any,
        a16: Any,
    ) -> Any:
        return f(a1)(a2)(a3)(a4)(a5)(a6)(a7)(a8)(a9)(a10)(a11)(a12)(a13)(a14)(a15)(a16)

    _curried[f2] = f
    return f2


def curry16(
    f: Callable[
        [
            _T1,
            _T2,
            _T3,
            _T4,
            _T5,
            _T6,
            _T7,
            _T8,
            _T9,
            _T10,
            _T11,
            _T12,
            _T13,
            _T14,
            _T15,
            _T16,
        ],
        _TResult,
    ],
) -> Callable[
    [_T1],
    Callable[
        [_T2],
        Callable[
            [_T3],
            Callable[
                [_T4],
                Callable[
                    [_T5],
                    Callable[
                        [_T6],
                        Callable[
                            [_T7],
                            Callable[
                                [_T8],
                                Callable[
                                    [_T9],
                                    Callable[
                                        [_T10],
                                        Callable[
                                            [_T11],
                                            Callable[
                                                [_T12],
                                                Callable[
                                                    [_T13],
                                                    Callable[
                                                        [_T14],
                                                        Callable[
                                                            [_T15],
                                                            Callable[[_T16], _TResult],
                                                        ],
                                                    ],
                                                ],
                                            ],
                                        ],
                                    ],
                                ],
                            ],
                        ],
                    ],
                ],
            ],
        ],
    ],
]:
    f2 = _curried.get(f)
    if f2 is None:
        return (
            lambda a1: lambda a2: lambda a3: lambda a4: lambda a5: lambda a6: lambda a7: lambda a8: lambda a9: lambda a10: lambda a11: lambda a12: lambda a13: lambda a14: lambda a15: lambda a16: f(
                a1, a2, a3, a4, a5, a6, a7, a8, a9, a10, a11, a12, a13, a14, a15, a16
            )
        )
    else:
        return f2


def uncurry17(
    f: Callable[
        [_T1],
        Callable[
            [_T2],
            Callable[
                [_T3],
                Callable[
                    [_T4],
                    Callable[
                        [_T5],
                        Callable[
                            [_T6],
                            Callable[
                                [_T7],
                                Callable[
                                    [_T8],
                                    Callable[
                                        [_T9],
                                        Callable[
                                            [_T10],
                                            Callable[
                                                [_T11],
                                                Callable[
                                                    [_T12],
                                                    Callable[
                                                        [_T13],
                                                        Callable[
                                                            [_T14],
                                                            Callable[
                                                                [_T15],
                                                                Callable[
                                                                    [_T16],
                                                                    Callable[[_T17], _TResult],
                                                                ],
                                                            ],
                                                        ],
                                                    ],
                                                ],
                                            ],
                                        ],
                                    ],
                                ],
                            ],
                        ],
                    ],
                ],
            ],
        ],
    ],
) -> Callable[
    [
        _T1,
        _T2,
        _T3,
        _T4,
        _T5,
        _T6,
        _T7,
        _T8,
        _T9,
        _T10,
        _T11,
        _T12,
        _T13,
        _T14,
        _T15,
        _T16,
        _T17,
    ],
    _TResult,
]:
    def f2(
        a1: Any,
        a2: Any,
        a3: Any,
        a4: Any,
        a5: Any,
        a6: Any,
        a7: Any,
        a8: Any,
        a9: Any,
        a10: Any,
        a11: Any,
        a12: Any,
        a13: Any,
        a14: Any,
        a15: Any,
        a16: Any,
        a17: Any,
    ) -> Any:
        return f(a1)(a2)(a3)(a4)(a5)(a6)(a7)(a8)(a9)(a10)(a11)(a12)(a13)(a14)(a15)(a16)(a17)

    _curried[f2] = f
    return f2


def curry17(
    f: Callable[
        [
            _T1,
            _T2,
            _T3,
            _T4,
            _T5,
            _T6,
            _T7,
            _T8,
            _T9,
            _T10,
            _T11,
            _T12,
            _T13,
            _T14,
            _T15,
            _T16,
            _T17,
        ],
        _TResult,
    ],
) -> Callable[
    [_T1],
    Callable[
        [_T2],
        Callable[
            [_T3],
            Callable[
                [_T4],
                Callable[
                    [_T5],
                    Callable[
                        [_T6],
                        Callable[
                            [_T7],
                            Callable[
                                [_T8],
                                Callable[
                                    [_T9],
                                    Callable[
                                        [_T10],
                                        Callable[
                                            [_T11],
                                            Callable[
                                                [_T12],
                                                Callable[
                                                    [_T13],
                                                    Callable[
                                                        [_T14],
                                                        Callable[
                                                            [_T15],
                                                            Callable[
                                                                [_T16],
                                                                Callable[[_T17], _TResult],
                                                            ],
                                                        ],
                                                    ],
                                                ],
                                            ],
                                        ],
                                    ],
                                ],
                            ],
                        ],
                    ],
                ],
            ],
        ],
    ],
]:
    f2 = _curried.get(f)
    if f2 is None:
        return (
            lambda a1: lambda a2: lambda a3: lambda a4: lambda a5: lambda a6: lambda a7: lambda a8: lambda a9: lambda a10: lambda a11: lambda a12: lambda a13: lambda a14: lambda a15: lambda a16: lambda a17: f(
                a1, a2, a3, a4, a5, a6, a7, a8, a9, a10, a11, a12, a13, a14, a15, a16, a17
            )
        )
    else:
        return f2


def uncurry18(
    f: Callable[
        [_T1],
        Callable[
            [_T2],
            Callable[
                [_T3],
                Callable[
                    [_T4],
                    Callable[
                        [_T5],
                        Callable[
                            [_T6],
                            Callable[
                                [_T7],
                                Callable[
                                    [_T8],
                                    Callable[
                                        [_T9],
                                        Callable[
                                            [_T10],
                                            Callable[
                                                [_T11],
                                                Callable[
                                                    [_T12],
                                                    Callable[
                                                        [_T13],
                                                        Callable[
                                                            [_T14],
                                                            Callable[
                                                                [_T15],
                                                                Callable[
                                                                    [_T16],
                                                                    Callable[
                                                                        [_T17],
                                                                        Callable[
                                                                            [_T18],
                                                                            _TResult,
                                                                        ],
                                                                    ],
                                                                ],
                                                            ],
                                                        ],
                                                    ],
                                                ],
                                            ],
                                        ],
                                    ],
                                ],
                            ],
                        ],
                    ],
                ],
            ],
        ],
    ],
) -> Callable[
    [
        _T1,
        _T2,
        _T3,
        _T4,
        _T5,
        _T6,
        _T7,
        _T8,
        _T9,
        _T10,
        _T11,
        _T12,
        _T13,
        _T14,
        _T15,
        _T16,
        _T17,
        _T18,
    ],
    _TResult,
]:
    def f2(
        a1: Any,
        a2: Any,
        a3: Any,
        a4: Any,
        a5: Any,
        a6: Any,
        a7: Any,
        a8: Any,
        a9: Any,
        a10: Any,
        a11: Any,
        a12: Any,
        a13: Any,
        a14: Any,
        a15: Any,
        a16: Any,
        a17: Any,
        a18: Any,
    ) -> Any:
        return f(a1)(a2)(a3)(a4)(a5)(a6)(a7)(a8)(a9)(a10)(a11)(a12)(a13)(a14)(a15)(a16)(a17)(a18)

    _curried[f2] = f
    return f2


def curry18(
    f: Callable[
        [
            _T1,
            _T2,
            _T3,
            _T4,
            _T5,
            _T6,
            _T7,
            _T8,
            _T9,
            _T10,
            _T11,
            _T12,
            _T13,
            _T14,
            _T15,
            _T16,
            _T17,
            _T18,
        ],
        _TResult,
    ],
) -> Callable[
    [_T1],
    Callable[
        [_T2],
        Callable[
            [_T3],
            Callable[
                [_T4],
                Callable[
                    [_T5],
                    Callable[
                        [_T6],
                        Callable[
                            [_T7],
                            Callable[
                                [_T8],
                                Callable[
                                    [_T9],
                                    Callable[
                                        [_T10],
                                        Callable[
                                            [_T11],
                                            Callable[
                                                [_T12],
                                                Callable[
                                                    [_T13],
                                                    Callable[
                                                        [_T14],
                                                        Callable[
                                                            [_T15],
                                                            Callable[
                                                                [_T16],
                                                                Callable[
                                                                    [_T17],
                                                                    Callable[[_T18], _TResult],
                                                                ],
                                                            ],
                                                        ],
                                                    ],
                                                ],
                                            ],
                                        ],
                                    ],
                                ],
                            ],
                        ],
                    ],
                ],
            ],
        ],
    ],
]:
    f2 = _curried.get(f)
    if f2 is None:
        return (
            lambda a1: lambda a2: lambda a3: lambda a4: lambda a5: lambda a6: lambda a7: lambda a8: lambda a9: lambda a10: lambda a11: lambda a12: lambda a13: lambda a14: lambda a15: lambda a16: lambda a17: lambda a18: f(
                a1,
                a2,
                a3,
                a4,
                a5,
                a6,
                a7,
                a8,
                a9,
                a10,
                a11,
                a12,
                a13,
                a14,
                a15,
                a16,
                a17,
                a18,
            )
        )
    else:
        return f2


def uncurry19(
    f: Callable[
        [_T1],
        Callable[
            [_T2],
            Callable[
                [_T3],
                Callable[
                    [_T4],
                    Callable[
                        [_T5],
                        Callable[
                            [_T6],
                            Callable[
                                [_T7],
                                Callable[
                                    [_T8],
                                    Callable[
                                        [_T9],
                                        Callable[
                                            [_T10],
                                            Callable[
                                                [_T11],
                                                Callable[
                                                    [_T12],
                                                    Callable[
                                                        [_T13],
                                                        Callable[
                                                            [_T14],
                                                            Callable[
                                                                [_T15],
                                                                Callable[
                                                                    [_T16],
                                                                    Callable[
                                                                        [_T17],
                                                                        Callable[
                                                                            [_T18],
                                                                            Callable[
                                                                                [_T19],
                                                                                _TResult,
                                                                            ],
                                                                        ],
                                                                    ],
                                                                ],
                                                            ],
                                                        ],
                                                    ],
                                                ],
                                            ],
                                        ],
                                    ],
                                ],
                            ],
                        ],
                    ],
                ],
            ],
        ],
    ],
) -> Callable[
    [
        _T1,
        _T2,
        _T3,
        _T4,
        _T5,
        _T6,
        _T7,
        _T8,
        _T9,
        _T10,
        _T11,
        _T12,
        _T13,
        _T14,
        _T15,
        _T16,
        _T17,
        _T18,
        _T19,
    ],
    _TResult,
]:
    def f2(
        a1: Any,
        a2: Any,
        a3: Any,
        a4: Any,
        a5: Any,
        a6: Any,
        a7: Any,
        a8: Any,
        a9: Any,
        a10: Any,
        a11: Any,
        a12: Any,
        a13: Any,
        a14: Any,
        a15: Any,
        a16: Any,
        a17: Any,
        a18: Any,
        a19: Any,
    ) -> Any:
        return f(a1)(a2)(a3)(a4)(a5)(a6)(a7)(a8)(a9)(a10)(a11)(a12)(a13)(a14)(a15)(a16)(a17)(a18)(a19)

    _curried[f2] = f
    return f2


def curry19(
    f: Callable[
        [
            _T1,
            _T2,
            _T3,
            _T4,
            _T5,
            _T6,
            _T7,
            _T8,
            _T9,
            _T10,
            _T11,
            _T12,
            _T13,
            _T14,
            _T15,
            _T16,
            _T17,
            _T18,
            _T19,
        ],
        _TResult,
    ],
) -> Callable[
    [_T1],
    Callable[
        [_T2],
        Callable[
            [_T3],
            Callable[
                [_T4],
                Callable[
                    [_T5],
                    Callable[
                        [_T6],
                        Callable[
                            [_T7],
                            Callable[
                                [_T8],
                                Callable[
                                    [_T9],
                                    Callable[
                                        [_T10],
                                        Callable[
                                            [_T11],
                                            Callable[
                                                [_T12],
                                                Callable[
                                                    [_T13],
                                                    Callable[
                                                        [_T14],
                                                        Callable[
                                                            [_T15],
                                                            Callable[
                                                                [_T16],
                                                                Callable[
                                                                    [_T17],
                                                                    Callable[
                                                                        [_T18],
                                                                        Callable[
                                                                            [_T19],
                                                                            _TResult,
                                                                        ],
                                                                    ],
                                                                ],
                                                            ],
                                                        ],
                                                    ],
                                                ],
                                            ],
                                        ],
                                    ],
                                ],
                            ],
                        ],
                    ],
                ],
            ],
        ],
    ],
]:
    f2 = _curried.get(f)
    if f2 is None:
        return (
            lambda a1: lambda a2: lambda a3: lambda a4: lambda a5: lambda a6: lambda a7: lambda a8: lambda a9: lambda a10: lambda a11: lambda a12: lambda a13: lambda a14: lambda a15: lambda a16: lambda a17: lambda a18: lambda a19: f(
                a1,
                a2,
                a3,
                a4,
                a5,
                a6,
                a7,
                a8,
                a9,
                a10,
                a11,
                a12,
                a13,
                a14,
                a15,
                a16,
                a17,
                a18,
                a19,
            )
        )
    else:
        return f2


def uncurry20(
    f: Callable[
        [_T1],
        Callable[
            [_T2],
            Callable[
                [_T3],
                Callable[
                    [_T4],
                    Callable[
                        [_T5],
                        Callable[
                            [_T6],
                            Callable[
                                [_T7],
                                Callable[
                                    [_T8],
                                    Callable[
                                        [_T9],
                                        Callable[
                                            [_T10],
                                            Callable[
                                                [_T11],
                                                Callable[
                                                    [_T12],
                                                    Callable[
                                                        [_T13],
                                                        Callable[
                                                            [_T14],
                                                            Callable[
                                                                [_T15],
                                                                Callable[
                                                                    [_T16],
                                                                    Callable[
                                                                        [_T17],
                                                                        Callable[
                                                                            [_T18],
                                                                            Callable[
                                                                                [_T19],
                                                                                Callable[
                                                                                    [_T20],
                                                                                    _TResult,
                                                                                ],
                                                                            ],
                                                                        ],
                                                                    ],
                                                                ],
                                                            ],
                                                        ],
                                                    ],
                                                ],
                                            ],
                                        ],
                                    ],
                                ],
                            ],
                        ],
                    ],
                ],
            ],
        ],
    ],
) -> Callable[
    [
        _T1,
        _T2,
        _T3,
        _T4,
        _T5,
        _T6,
        _T7,
        _T8,
        _T9,
        _T10,
        _T11,
        _T12,
        _T13,
        _T14,
        _T15,
        _T16,
        _T17,
        _T18,
        _T19,
        _T20,
    ],
    _TResult,
]:
    def f2(
        a1: Any,
        a2: Any,
        a3: Any,
        a4: Any,
        a5: Any,
        a6: Any,
        a7: Any,
        a8: Any,
        a9: Any,
        a10: Any,
        a11: Any,
        a12: Any,
        a13: Any,
        a14: Any,
        a15: Any,
        a16: Any,
        a17: Any,
        a18: Any,
        a19: Any,
        a20: Any,
    ) -> Any:
        return f(a1)(a2)(a3)(a4)(a5)(a6)(a7)(a8)(a9)(a10)(a11)(a12)(a13)(a14)(a15)(a16)(a17)(a18)(a19)(a20)

    _curried[f2] = f
    return f2


def curry20(
    f: Callable[
        [
            _T1,
            _T2,
            _T3,
            _T4,
            _T5,
            _T6,
            _T7,
            _T8,
            _T9,
            _T10,
            _T11,
            _T12,
            _T13,
            _T14,
            _T15,
            _T16,
            _T17,
            _T18,
            _T19,
            _T20,
        ],
        _TResult,
    ],
) -> Callable[
    [_T1],
    Callable[
        [_T2],
        Callable[
            [_T3],
            Callable[
                [_T4],
                Callable[
                    [_T5],
                    Callable[
                        [_T6],
                        Callable[
                            [_T7],
                            Callable[
                                [_T8],
                                Callable[
                                    [_T9],
                                    Callable[
                                        [_T10],
                                        Callable[
                                            [_T11],
                                            Callable[
                                                [_T12],
                                                Callable[
                                                    [_T13],
                                                    Callable[
                                                        [_T14],
                                                        Callable[
                                                            [_T15],
                                                            Callable[
                                                                [_T16],
                                                                Callable[
                                                                    [_T17],
                                                                    Callable[
                                                                        [_T18],
                                                                        Callable[
                                                                            [_T19],
                                                                            Callable[
                                                                                [_T20],
                                                                                _TResult,
                                                                            ],
                                                                        ],
                                                                    ],
                                                                ],
                                                            ],
                                                        ],
                                                    ],
                                                ],
                                            ],
                                        ],
                                    ],
                                ],
                            ],
                        ],
                    ],
                ],
            ],
        ],
    ],
]:
    f2 = _curried.get(f)
    if f2 is None:
        return (
            lambda a1: lambda a2: lambda a3: lambda a4: lambda a5: lambda a6: lambda a7: lambda a8: lambda a9: lambda a10: lambda a11: lambda a12: lambda a13: lambda a14: lambda a15: lambda a16: lambda a17: lambda a18: lambda a19: lambda a20: f(
                a1,
                a2,
                a3,
                a4,
                a5,
                a6,
                a7,
                a8,
                a9,
                a10,
                a11,
                a12,
                a13,
                a14,
                a15,
                a16,
                a17,
                a18,
                a19,
                a20,
            )
        )
    else:
        return f2


def is_array_like(x: Any) -> bool:
    return isinstance(x, list | tuple | set | array | bytes | bytearray)


def is_disposable(x: Any) -> bool:
    return x is not None and isinstance(x, IDisposable)


def dispose(x: Disposable | AbstractContextManager[Any]) -> None:
    """Helper to dispose objects.

    Also tries to call `__exit__` if the object turns out to be a Python resource manager.
    For more info see: https://www.python.org/dev/peps/pep-0310/
    """
    try:
        x.Dispose()  # type: ignore
    except AttributeError as ex:
        try:
            x.__exit__(None, None, None)
        except AttributeError:
            raise ex


def is_hashable(x: Any) -> bool:
    return hasattr(x, "GetHashCode")


def is_hashable_py(x: Any) -> bool:
    return hasattr(x, "__hash__") and callable(x.__hash__)


def to_iterator(en: IEnumerator[_T]) -> IEnumerator[_T]:
    class Iterator:
        def __iter__(self):
            return self

        def __next__(self):
            has_next = en.System_Collections_IEnumerator_MoveNext()
            if not has_next:
                raise StopIteration
            return en.System_Collections_IEnumerator_get_Current()

    return Enumerator(Iterator())


class ObjectRef:
    id_map: ClassVar = dict[int, int]()
    count: ClassVar = 0

    @staticmethod
    def id(o: Any) -> int:
        _id = id(o)
        if _id not in ObjectRef.id_map:
            ObjectRef.count += 1
            ObjectRef.id_map[_id] = ObjectRef.count

        return ObjectRef.id_map[_id]


def safe_hash(x: Any) -> int:
    return (
        0
        if x is None
        else hash(x)
        if is_hashable_py(x)
        else x.GetHashCode()
        if is_hashable(x)
        else number_hash(ObjectRef.id(x))
    )


def string_hash(s: str) -> int:
    h = 5381
    for c in s:
        h = (h * 33) ^ ord(c)

    return h


def number_hash(x: int) -> int:
    return x * 2654435761 | 0


def identity_hash(x: Any) -> int:
    if x is None:
        return 0

    if is_hashable(x):
        return x.GetHashCode()

    if is_hashable_py(x):
        return hash(x)

    return physical_hash(x)


def combine_hash_codes(hashes: list[int]) -> int:
    if not hashes:
        return 0

    return functools.reduce(lambda h1, h2: ((h1 << 5) + h1) ^ h2, hashes)


def structural_hash(x: Any) -> int:
    return hash(x)


def array_hash(xs: list[Any]) -> int:
    hashes: list[int] = []
    for x in xs:
        hashes.append(structural_hash(x))

    return combine_hash_codes(hashes)


def physical_hash(x: Any) -> int:
    if hasattr(x, "__hash__") and callable(x.__hash__):
        return hash(x)

    return number_hash(ObjectRef.id(x))


def round(value: float, digits: int = 0):
    m = pow(10, digits)
    n = +(value * m if digits else value)
    i = math.floor(n)
    f = n - i
    e = 1e-8
    r = (i if (i % 2 == 0) else i + 1) if (f > 0.5 - e and f < 0.5 + e) else builtins.round(n)
    return r / m if digits else r


def randint(a: int, b: int) -> int:
    return random.randint(a, b - 1)


def unescape_data_string(s: str) -> str:
    # https://stackoverflow.com/a/4458580/524236
    return unquote(re.sub(r"\+", "%20", s))


def escape_data_string(s: str) -> str:
    return quote(s, safe="")


def escape_uri_string(s: str) -> str:
    return quote(s, safe="&?:/!=")


def ignore(a: Any = None) -> None:
    """Ignore argument, returns None."""
    return


def copy_to_array(src: Array[_T], srci: int, trg: Array[_T], trgi: int, cnt: int) -> None:
    for i in range(0, cnt, 1):
        trg[trgi + i] = src[srci + i]


class PlatformID(IntEnum):
    Win32S = 0
    Win32Windows = 1
    Win32NT = 2
    WinCE = 3
    Unix = 4
    Xbox = 5
    MacOSX = 6
    Other = 7


def get_platform() -> PlatformID:
    name = platform.platform(terse=True).lower()

    if name.startswith("windows"):
        return PlatformID.Win32NT
    elif name.startswith("linux"):
        return PlatformID.Unix
    elif name.startswith("macos"):
        return PlatformID.MacOSX

    return PlatformID.Other



from __future__ import annotations
from collections.abc import Callable
from typing import (Any, TypeVar)
from ...arctrl_core.Helper.collections_ import ResizeArray_map
from ...arctrl_core.ontology_annotation import OntologyAnnotation
from ...arctrl_core.Process.component import (Component_fromISAString_7C9A7CF8, Component, Component_toStringObject_Z685B8F25)
from ...arctrl_core.Process.protocol_parameter import ProtocolParameter
from ...fable_library.array_ import (fold, map3, fill, equals_with)
from ...fable_library.list import (empty, of_array, FSharpList, fold as fold_1, map)
from ...fable_library.option import (some, value, default_arg)
from ...fable_library.string_ import (to_fail, printf, to_text)
from ...fable_library.types import Array
from ...fable_library.util import equals
from ..collection_aux import Array_map4

_T = TypeVar("_T")

def Option_fromValueWithDefault(d: Any, v: Any) -> Any | None:
    if equals(d, v):
        return None

    else: 
        return some(v)



def Option_mapDefault(d: Any, f: Callable[[_T], _T], o: Any | None=None) -> Any | None:
    return Option_fromValueWithDefault(d, f(d) if (o is None) else f(value(o)))


def OntologyAnnotation_getLengthOfAggregatedStrings(separator: str, strings: Array[str]) -> int:
    def folder(l: int, s: str, separator: Any=separator, strings: Any=strings) -> int:
        if s == "":
            return l

        elif l == 0:
            return len(s.split(separator))

        elif l == len(s.split(separator)):
            return l

        else: 
            return to_fail(printf("The length of the aggregated string %s does not match the length of the others"))(s)


    return fold(folder, 0, strings)


def OntologyAnnotation_fromAggregatedStrings(separator: str, terms: str, source: str, accessions: str) -> Array[OntologyAnnotation]:
    l: int = OntologyAnnotation_getLengthOfAggregatedStrings(separator, [terms, source, accessions]) or 0
    if l == 0:
        return []

    else: 
        def _arrow3421(a: str, b: str, c: str, separator: Any=separator, terms: Any=terms, source: Any=source, accessions: Any=accessions) -> OntologyAnnotation:
            return OntologyAnnotation.create(a, b, c)

        return list(map3(_arrow3421, fill([0] * l, 0, l, "") if (terms == "") else terms.split(separator), fill([0] * l, 0, l, "") if (source == "") else source.split(separator), fill([0] * l, 0, l, "") if (accessions == "") else accessions.split(separator), None))



def OntologyAnnotation_toAggregatedStrings(separator: str, oas: Array[OntologyAnnotation]) -> dict[str, Any]:
    first: bool = True
    def _arrow3422(x: OntologyAnnotation, y: OntologyAnnotation, separator: Any=separator, oas: Any=oas) -> bool:
        return equals(x, y)

    if equals_with(_arrow3422, oas, []):
        return {
            "TermAccessionNumberAgg": "",
            "TermNameAgg": "",
            "TermSourceREFAgg": ""
        }

    else: 
        def folder(tupled_arg: tuple[str, str, str], term: OntologyAnnotation, separator: Any=separator, oas: Any=oas) -> tuple[str, str, str]:
            nonlocal first
            match_value: str = default_arg(term.Name, "")
            tsr: str = default_arg(term.TermSourceREF, "")
            tan: str = default_arg(term.TermAccessionNumber, "")
            name: str = match_value
            if first:
                first = False
                return (name, tsr, tan)

            else: 
                return (to_text(printf("%s%c%s"))(tupled_arg[0])(separator)(name), to_text(printf("%s%c%s"))(tupled_arg[1])(separator)(tsr), to_text(printf("%s%c%s"))(tupled_arg[2])(separator)(tan))


        tupled_arg_1: tuple[str, str, str] = fold(folder, ("", "", ""), oas)
        return {
            "TermAccessionNumberAgg": tupled_arg_1[2],
            "TermNameAgg": tupled_arg_1[0],
            "TermSourceREFAgg": tupled_arg_1[1]
        }



def Component_fromAggregatedStrings(separator: str, names: str, terms: str, source: str, accessions: str) -> FSharpList[Component]:
    l: int = OntologyAnnotation_getLengthOfAggregatedStrings(separator, [names, terms, source, accessions]) or 0
    if l == 0:
        return empty()

    else: 
        def _arrow3423(a: str, b: str, c: str, d: str, separator: Any=separator, names: Any=names, terms: Any=terms, source: Any=source, accessions: Any=accessions) -> Component:
            return Component_fromISAString_7C9A7CF8(a, b, c, d)

        return of_array(Array_map4(_arrow3423, fill([0] * l, 0, l, "") if (names == "") else names.split(separator), fill([0] * l, 0, l, "") if (terms == "") else terms.split(separator), fill([0] * l, 0, l, "") if (source == "") else source.split(separator), fill([0] * l, 0, l, "") if (accessions == "") else accessions.split(separator)))



def Component_toAggregatedStrings(separator: str, cs: FSharpList[Component]) -> dict[str, Any]:
    first: bool = True
    if equals(cs, empty()):
        return {
            "NameAgg": "",
            "TermAccessionNumberAgg": "",
            "TermNameAgg": "",
            "TermSourceREFAgg": ""
        }

    else: 
        def folder(tupled_arg: tuple[str, str, str, str], tupled_arg_1: tuple[str, dict[str, Any]], separator: Any=separator, cs: Any=cs) -> tuple[str, str, str, str]:
            nonlocal first
            name: str = tupled_arg_1[0]
            term: dict[str, Any] = tupled_arg_1[1]
            if first:
                first = False
                return (name, term["TermName"], term["TermSourceREF"], term["TermAccessionNumber"])

            else: 
                return (to_text(printf("%s%c%s"))(tupled_arg[0])(separator)(name), to_text(printf("%s%c%s"))(tupled_arg[1])(separator)(term["TermName"]), to_text(printf("%s%c%s"))(tupled_arg[2])(separator)(term["TermSourceREF"]), to_text(printf("%s%c%s"))(tupled_arg[3])(separator)(term["TermAccessionNumber"]))


        def mapping(c: Component, separator: Any=separator, cs: Any=cs) -> tuple[str, dict[str, Any]]:
            return Component_toStringObject_Z685B8F25(c)

        tupled_arg_2: tuple[str, str, str, str] = fold_1(folder, ("", "", "", ""), map(mapping, cs))
        return {
            "NameAgg": tupled_arg_2[0],
            "TermAccessionNumberAgg": tupled_arg_2[3],
            "TermNameAgg": tupled_arg_2[1],
            "TermSourceREFAgg": tupled_arg_2[2]
        }



def ProtocolParameter_fromAggregatedStrings(separator: str, terms: str, source: str, accessions: str) -> Array[ProtocolParameter]:
    def f(arg: OntologyAnnotation, separator: Any=separator, terms: Any=terms, source: Any=source, accessions: Any=accessions) -> ProtocolParameter:
        def f2(parameter_name: OntologyAnnotation | None=None, arg: Any=arg) -> ProtocolParameter:
            return ProtocolParameter.make(None, parameter_name)

        return f2(arg)

    return ResizeArray_map(f, OntologyAnnotation_fromAggregatedStrings(separator, terms, source, accessions))


def ProtocolParameter_toAggregatedStrings(separator: str, oas: FSharpList[ProtocolParameter]) -> dict[str, Any]:
    first: bool = True
    if equals(oas, empty()):
        return {
            "TermAccessionNumberAgg": "",
            "TermNameAgg": "",
            "TermSourceREFAgg": ""
        }

    else: 
        def folder(tupled_arg: tuple[str, str, str], term: dict[str, Any], separator: Any=separator, oas: Any=oas) -> tuple[str, str, str]:
            nonlocal first
            if first:
                first = False
                return (term["TermName"], term["TermSourceREF"], term["TermAccessionNumber"])

            else: 
                return (to_text(printf("%s%c%s"))(tupled_arg[0])(separator)(term["TermName"]), to_text(printf("%s%c%s"))(tupled_arg[1])(separator)(term["TermSourceREF"]), to_text(printf("%s%c%s"))(tupled_arg[2])(separator)(term["TermAccessionNumber"]))


        def mapping(pp: ProtocolParameter, separator: Any=separator, oas: Any=oas) -> dict[str, Any]:
            return ProtocolParameter.to_string_object(pp)

        tupled_arg_1: tuple[str, str, str] = fold_1(folder, ("", "", ""), map(mapping, oas))
        return {
            "TermAccessionNumberAgg": tupled_arg_1[2],
            "TermNameAgg": tupled_arg_1[0],
            "TermSourceREFAgg": tupled_arg_1[1]
        }



__all__ = ["Option_fromValueWithDefault", "Option_mapDefault", "OntologyAnnotation_getLengthOfAggregatedStrings", "OntologyAnnotation_fromAggregatedStrings", "OntologyAnnotation_toAggregatedStrings", "Component_fromAggregatedStrings", "Component_toAggregatedStrings", "ProtocolParameter_fromAggregatedStrings", "ProtocolParameter_toAggregatedStrings"]




from __future__ import annotations
from typing import Any
from ...arctrl_core.data import Data
from ...arctrl_core.Process.material import Material
from ...arctrl_core.Process.process_output import ProcessOutput
from ...arctrl_core.Process.sample import Sample
from ...fable_library.list import of_array
from ...thoth_json_core.decode import (one_of, map)
from ...thoth_json_core.types import (IEncodable, Decoder_1)
from ..data import (ROCrate_encoder as ROCrate_encoder_1, ROCrate_decoder as ROCrate_decoder_2, ISAJson_encoder as ISAJson_encoder_1, ISAJson_decoder as ISAJson_decoder_2)
from .material import (ROCrate_encoder as ROCrate_encoder_2, ROCrate_decoder as ROCrate_decoder_3, ISAJson_encoder as ISAJson_encoder_2, ISAJson_decoder as ISAJson_decoder_3)
from .sample import (ROCrate_encoder as ROCrate_encoder_3, ROCrate_decoder as ROCrate_decoder_1, ISAJson_encoder as ISAJson_encoder_3, ISAJson_decoder as ISAJson_decoder_1)

def ROCrate_encoder(value: ProcessOutput) -> IEncodable:
    if value.tag == 1:
        return ROCrate_encoder_1(value.fields[0])

    elif value.tag == 2:
        return ROCrate_encoder_2(value.fields[0])

    else: 
        return ROCrate_encoder_3(value.fields[0])



def _arrow2664(Item: Sample) -> ProcessOutput:
    return ProcessOutput(0, Item)


def _arrow2665(Item_1: Data) -> ProcessOutput:
    return ProcessOutput(1, Item_1)


def _arrow2666(Item_2: Material) -> ProcessOutput:
    return ProcessOutput(2, Item_2)


ROCrate_decoder: Decoder_1[ProcessOutput] = one_of(of_array([map(_arrow2664, ROCrate_decoder_1), map(_arrow2665, ROCrate_decoder_2), map(_arrow2666, ROCrate_decoder_3)]))

def ISAJson_encoder(id_map: Any | None, value: ProcessOutput) -> IEncodable:
    if value.tag == 1:
        return ISAJson_encoder_1(id_map, value.fields[0])

    elif value.tag == 2:
        return ISAJson_encoder_2(id_map, value.fields[0])

    else: 
        return ISAJson_encoder_3(id_map, value.fields[0])



def _arrow2668(Item: Sample) -> ProcessOutput:
    return ProcessOutput(0, Item)


def _arrow2669(Item_1: Data) -> ProcessOutput:
    return ProcessOutput(1, Item_1)


def _arrow2670(Item_2: Material) -> ProcessOutput:
    return ProcessOutput(2, Item_2)


ISAJson_decoder: Decoder_1[ProcessOutput] = one_of(of_array([map(_arrow2668, ISAJson_decoder_1), map(_arrow2669, ISAJson_decoder_2), map(_arrow2670, ISAJson_decoder_3)]))

__all__ = ["ROCrate_encoder", "ROCrate_decoder", "ISAJson_encoder", "ISAJson_decoder"]

